# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""This module implements the component model of the framework.

This model is inspired by the Seaside one. With the possibility to embed,
replace and call a component. It's described in `ComponentModel`
"""

import sys
import types

from peak.rules import when

from nagare import presentation, callbacks, continuation, partial

_marker = object()


class AnswerWithoutCall(BaseException):
    pass


class Component(object):
    """This class transforms any Python object into a component

    A component has views, can the embedded, replaced, called and can answsered a value.
    """
    def __init__(self, o=None, model=0, url=None):
        """Initialisation

        In:
          - ``o`` -- the python object (or component) to transform into a component
          - ``model`` -- the name of the view to use (``None`` for the default view)
          - ``url`` -- the url fragment to add before all the links generated by
            views of this component
        """
        self._becomes(o, model, url)

        self._cont = None
        self._on_answer = None

    def __call__(self):
        """Return the inner object
        """
        return self.o

    def register_callback(self, model, priority, callback, with_request, render):
        """Register a callback for this component

        In:
          - ``model`` -- name of the view which registers this callback (``None`` for the default view)
          - ``priority`` -- type and priority of the callback
          - ``callback`` -- the action function or method
          - ``with_request`` -- will the request and response objects be passed to the action?
          - ``render`` -- the render function or method
        """
        self.__dict__.setdefault('_new_callbacks', {})
        return callbacks.register(model, priority, callback, with_request, render, self._new_callbacks)

    def serialize_callbacks(self, clean_callbacks):
        """Return the callbacks to serialize

        In:
          - ``clean_callbacks`` -- do we have to forget all the old callbacks?

        Return:
          - the callbacks of this component
        """
        old = self.__dict__.pop('_callbacks', {})
        new = self.__dict__.pop('_new_callbacks', {})

        if not clean_callbacks:
            # Selectively keep some old callbacks

            # --
            # Policy #1: keep the old callbacks only if no view was rendered
            # => simple cleanup but can lead to ``CallbackLookup`` errors
            # if not new:
            #     new = old
            # --

            # --
            # Policy #2: no cleanup, always keep all the old callbacks
            # => no ``CallbackLookup`` errors but the sessions can keep growing
            #    in size as they are only reset when a new complete synchronous
            #    page rendering is done
            # new.update(old)
            # --

            # --
            # Policy #3: keep the old callbacks registered by a view only if
            # this view was not rendered again
            new.update(callbacks.clean(old, new))
            # --

        if new:
            self._callbacks = new

        return new

    def render(self, renderer, model=0):
        """Rendering method of a component

        Forward the call to the generic method of the ``presentation`` service
        """
        return presentation.render(self, renderer, self, model)

    def init(self, url, http_method, request):
        """Initialisation from an url

        Forward the call to the generic method of the ``presentation`` service
        """
        return presentation.init(self, url, self, http_method, request)

    def _becomes(self, o, model, url):
        """Replace a component by an object or an other component

        In:
          - ``o`` -- object to be replaced by
          - ``model`` -- the name of the view to use (``None`` for the default view)
          - ``url`` -- the url fragment to add before all the links generated by
            views of this component

        Return:
          - ``self``
        """
        o = self if o.__class__ is object else o
        if isinstance(o, Component):
            o = o()

        self.o = o

        self.model = model
        self.url = url

    def becomes(self, o=_marker, model=0, url=None):
        """Replace a component by an object or an other component

        In:
          - ``o`` -- object to be replaced by
          - ``model`` -- the name of the view to use (``None`` for the default view)
          - ``url`` -- the url fragment to add before all the links generated by
            views of this component

        Return:
          - ``self``
        """
        self._becomes(o, model, url or self.url)
        self._cont = None

        return self

    def call(self, o=_marker, model=0, url=None):
        """Call an other object or component

        The current component is replaced and will be back when the object
        will do an ``answer()``

        In:
          - ``o`` -- the object to call
          - ``model`` -- the name of the view to use (``None`` for the default view)
          - ``url`` -- the url fragment to add before all the links generated by
            views of this component

        Return:
          - the answer of the called object
        """
        sys.exc_clear()

        # Keep my configuration
        previous_o = self.o
        previous_model = self.model
        previous_url = self.url
        previous_cont = self._cont
        previous_on_answer = self._on_answer

        # Set the new configuration
        self._on_answer = None

        # Replace me by the object and wait its answer
        self.becomes(o, model, url)

        self._cont = continuation.get_current()

        try:
            r = self._cont.switch()
        except NotImplementedError:
            raise NotImplementedError('Stackless Python or PyPy is needed for `comp.call()`')

        # Restore my configuration
        self._on_answer = previous_on_answer
        self._cont = previous_cont
        self._becomes(previous_o, previous_model, previous_url)

        # Return the answer
        return r

    def answer(self, r=None):
        """Answer to a call

        In:
          - the value to answer
        """
        # If a function is listening to my answer, calls it
        if self._on_answer is not None:
            return self._on_answer(r)

        # Else, check if I was called by a component
        if self._cont is None:
            raise AnswerWithoutCall(self)

        # Returns my answer to the calling component
        self._cont.switch(r)

    def on_answer(self, f, *args, **kw):
        """
        Register a function to listen to my answer

        In:
          - ``f`` -- function to call with my answer
          - ``args``, ``kw`` -- ``f`` parameters
        """
        self._on_answer = partial.Partial(f, *args, **kw) if args or kw else f
        return self

    def __repr__(self):
        return '<%s with %s at 0x%x on object %r>' % (
            self.__class__.__name__.lower(),
            "model '%s'" % self.model if self.model else 'default model',
            id(self),
            self.o
        )


@when(presentation.render, (Component, object, object, int))
@when(presentation.render, (Component, object, object, types.NoneType))
@when(presentation.render, (Component, object, object, str))
def render(self, renderer, comp, model):
    """Rendering of a ``Component``

    In:
      - ``renderer`` -- the renderer
      - ``comp`` -- the component
      - ``model`` -- the name of the view

    Return:
      - the view of the component object
    """
    renderer = renderer.new()   # Create a new renderer of the same class than the current renderer
    renderer.start_rendering(self, model)

    if model == 0:
        model = self.model

    output = presentation.render(self(), renderer, self, model)
    return renderer.end_rendering(output)


@presentation.init_for(Component)
def init_for(self, url, comp, http_method, request):
    """Initialisation from an url

    In:
      - ``url`` -- rest of the url to process
      - ``comp`` -- the component
      - ``http_method`` -- the HTTP method
      - ``request`` -- the complete Request object

    Return:
      - ``presentation.NOT_FOUND`` if the url is invalid, else ``None``
    """
    presentation.init(self(), url, self, http_method, request)


# -----------------------------------------------------------------------------------------------------

class Task(object):
    """A ``Task`` encapsulated a simple method. A ``task`` is typically used to
    manage other components by calling them.

    .. warning::

       A ``Task`` is an object, not a component: you must wrap it into a ``Component()`` to use it.
    """

    def _go(self, comp):
        # If I was not called by an other component and nobody is listening to
        # my answer,  I'm the root component. So I call my ``go()`` method forever
        if comp._cont is comp._on_answer is None:
            while True:
                self.go(comp)

        # Else, answer with the return of the ``go`` method
        comp.answer(self.go(comp))

    def go(self, comp):
        raise BaseException('AbstractMethod')


@presentation.render_for(Task)
def render_task(self, renderer, comp, *args):
    return presentation.render(self._go, renderer, comp, *args)


# -----------------------------------------------------------------------------------------------------

@when(presentation.render, (types.FunctionType,))
@when(presentation.render, (types.MethodType,))
def render_function(f, renderer, comp, *args):
    continuation.Continuation(f, comp)
    return comp.render(renderer.parent)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Helper to mark an object as stateless"""

import random


def stateless(o):
    """Mark an object as stateless

    In:
      - ``o`` -- the object

    Return:
      - ``o``
    """
    if not hasattr(o, '_persistent_id'):
        # A stateless object has a unique oid
        o._persistent_id = random.randint(0, 99999999)

    return o


def stateful(o):
    """Mark an object as stateful

    In:
      - ``o`` -- the object

    Return:
      - ``o``
    """
    if hasattr(o, '_persistent_id'):
        del o._persistent_id

    return o



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""A LRU dictionary is a dictionary with a fixed maximum number of keys.

When this maximum is reached, the last recently used key is deleted when a new
key is added.
"""

from __future__ import with_statement

import threading


class LRUDict(object):
    """A LRU dictionary is a dictionary with a fixed maximum number of keys"""

    def __init__(self, size):
        """Initialization

        In:
          -  ``size`` -- maximum number of keys
        """
        self.size = size

        self.oldest = self.newest = 0  # Age of the oldest key / age of the last recently used key
        self.age_to_items = {}         # Dict: key_age -> key
        self.items = {}                # Dict: key -> (key_age, value)

    def __contains__(self, k):
        """Test if a key exists into this dictionary

        In:
          -  ``k`` -- the key

        Return:
          - a boolean
        """
        return k in self.items

    def _set_newest(self, k, o):
        """Insert a key as the last recently used

        In:
           - ``k`` -- the key
           - ``o`` -- the value
        """
        self.age_to_items[self.newest] = k
        self.items[k] = (self.newest, o)
        self.newest += 1

    def __getitem__(self, k):
        """Return the value of a key.

        The key becomes the most recently used key.

        In:
          - ``k`` -- the key

        Return:
          - the value
        """
        (age, item) = self.items[k]
        del self.age_to_items[age]

        self._set_newest(k, item)
        return item

    def __setitem__(self, k, o):
        """Set the value of a key.

        The key becomes the most recently used key.

        In:
          - ``k`` -- the key
          - ``o`` -- the value
        """
        try:
            (age, item) = self.items[k]
            del self.age_to_items[age]
        except KeyError:
            pass

        self._set_newest(k, o)

        if len(self.items) > self.size:
            # Maximum number of keys reached
            # Delete the last recently used key
            while self.oldest not in self.age_to_items:
                self.oldest += 1

            del self.items[self.age_to_items.pop(self.oldest)]

    def __delitem__(self, k):
        """Delete a key.

        In:
          - ``k`` -- the key
        """
        (age, item) = self.items.pop(k)
        del self.age_to_items[age]

    def debug(self):
        print self.oldest, self.newest, self.age_to_items, self.items


class ThreadSafeLRUDict(LRUDict):
    """Tread safe version of a LRU dictionary"""

    def __init__(self, *args, **kw):
        super(ThreadSafeLRUDict, self).__init__(*args, **kw)
        self.lock = threading.RLock()

    def __contains__(self, k):
        """Test if a key exists into this dictionary

        In:
          -  ``k`` -- the key

        Return:
          - a boolean
        """
        with self.lock:
            return k in self.items

    def __getitem__(self, k):
        with self.lock:
            return super(ThreadSafeLRUDict, self).__getitem__(k)

    def __setitem__(self, k, o):
        with self.lock:
            super(ThreadSafeLRUDict, self).__setitem__(k, o)


# ----------------------------------------------------------------------------

if __name__ == '__main__':
    cache = ThreadSafeLRUDict(3)
    cache['a'] = 1
    cache.debug()
    cache['b'] = 2
    cache.debug()
    cache['a'] = 3
    cache.debug()
    cache['a'] = 4
    cache.debug()
    cache['b'] = 5
    cache.debug()
    x = cache['a']
    cache.debug()

    cache['c'] = 'c'
    cache.debug()

    cache['d'] = 'd'
    cache.debug()

    cache['b'] = 'b'
    cache.debug()

    cache['e'] = 'e'
    cache.debug()



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Framework environment administrative view
"""

import sys
import time
import pkg_resources

from nagare import presentation


class Admin(object):
    priority = 1        # Order of the default view, into the administrative interface

    def __init__(self, apps):
        """Initialisation

        In:
          - ``apps`` -- list of tuples (application, application name, application urls)
        """
        pass


@presentation.render_for(Admin)
def render(self, h, *args):
    """Display informations about the running framework environment"""
    version = sys.version.split('\n')

    return h.div(
        h.h2('Informations'),

        h.p('Nagare - version %s - %s -' % (pkg_resources.get_distribution('nagare').version, time.strftime('%c'))),
        # h.p('Installed options: ', ', '.join(pkg_resources.get_distribution('fw-ng').extras)),

        h.p(sys.subversion[0], ' ', version[0]),
        [h.p(v) for v in version[1:]]
    )



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Internationalization service
"""

import os
import datetime

import pkg_resources
from peak.rules import when

from nagare.namespaces import xml
from nagare import serializer, ajax, local, log

try:
    from babel import core, negotiate_locale, Locale as CoreLocale
    from babel import dates, numbers, support
    import pytz

    class LazyProxy(support.LazyProxy):
        """Picklable ``babel.support.LazyProxy`` objects
        """
        @property
        def value(self):
            """Always evaluate, without any cache
            """
            return self._func(*self._args, **self._kwargs)

        def __getstate__(self):
            return self._func, self._args, self._kwargs

        def __setstate__(self, attrs):
            self.__init__(attrs[0], *attrs[1], **attrs[2])

    @when(xml.add_child, (xml._Tag, LazyProxy))
    def add_child(self, lazy):
        """Add a lazy string to a tag

        In:
          - ``self`` -- the tag
          - ``lazy`` -- the lazy string to add

        """
        self.add_child(lazy.value)

    @when(ajax.py2js, (LazyProxy,))
    def py2js(lazy, h):
        """Generic method to transcode a lazy string to javascript

        In:
          - ``lazy`` -- the lazy string
          - ``h`` -- the current renderer

        Return:
          - transcoded javascript
        """

        return ajax.py2js(lazy.value, h)

    @when(serializer.serialize, (LazyProxy,))
    def serialize(lazy, content_type, doctype, declaration):
        """Generic method to generate a text from a lazy string

        In:
          - ``lazy`` -- the lazy string
          - ``content_type`` -- the rendered content type
          - ``doctype`` -- the (optional) doctype
          - ``declaration`` -- is the XML declaration to be outputed ?

        Return:
          - a tuple (content_type, content)
        """
        return serializer.serialize(lazy.value, content_type, doctype, declaration)

except ImportError:

    # Dummy ``Locale`` base class and ``LazyProxy``, used when the i18n extra
    # is not installed

    class CoreLocale(object):
        def __init__(self, language, territory, script, variant):
            if not (language is territory is script is variant is None):
                log.warning('i18n extra not installed')

            self.language = None
            self.zone_formats = {'region': '%s'}

    def LazyProxy(f, *args, **kw):
        return f(*args, **kw)


# -----------------------------------------------------------------------------

# Mac Leopard OS bug (see http://code.djangoproject.com/ticket/5846)

if os.environ.get('LC_CTYPE', '').lower() == 'utf-8':
    os.environ['LC_CTYPE'] = 'en_US.utf-8'


# -----------------------------------------------------------------------------

# Service API
# -----------

def gettext(msg, domain=None, **kw):
    return get_locale().gettext(msg, domain, **kw)


def ugettext(msg, domain=None, **kw):
    return get_locale().ugettext(msg, domain, **kw)
_ = ugettext  # noqa: E305


def ngettext(singular, plural, n, domain=None, **kw):
    return get_locale().ngettext(singular, plural, n, domain, **kw)


def ungettext(singular, plural, n, domain=None, **kw):
    return get_locale().ungettext(singular, plural, n, domain, **kw)
_N = ungettext  # noqa: E305


def lazy_gettext(msg, domain=None, **kw):
    return LazyProxy(gettext, msg, domain, **kw)


def lazy_ugettext(msg, domain=None, **kw):
    return LazyProxy(ugettext, msg, domain, **kw)
_L = lazy_ugettext  # noqa: E305


def lazy_ngettext(singular, plural, n, domain=None, **kw):
    return LazyProxy(ngettext, singular, plural, n, domain, **kw)


def lazy_ungettext(singular, plural, n, domain=None, **kw):
    return LazyProxy(ungettext, singular, plural, n, domain, **kw)
_LN = lazy_ungettext  # noqa: E305


def get_period_names():
    return get_locale().get_period_names()


def get_day_names(width='wide', context='format'):
    return get_locale().get_day_names(width, context)


def get_month_names(width='wide', context='format'):
    return get_locale().get_month_names(width, context)


def get_quarter_names(width='wide', context='format'):
    return get_locale().get_quarter_names(width, context)


def get_era_names(width='wide'):
    return get_locale().get_era_names(width)


def get_date_format(format='medium'):
    return get_locale().get_date_format(format)


def get_datetime_format(format='medium'):
    return get_locale().get_datetime_format(format)


def get_time_format(format='medium'):
    return get_locale().get_time_format(format)


def get_timezone_gmt(dt=None, width='long'):
    return get_locale().get_timezone_gmt(dt, width)


def get_timezone_location(dt_or_timezone=None):
    return get_locale().get_timezone_location(dt_or_timezone)


def get_timezone_name(dt_or_timezone=None, width='long', uncommon=False):
    return get_locale().get_timezone_name(dt_or_timezone, width, uncommon)


def get_currency_name(currency):
    return get_locale().get_currency_name(currency)


def get_currency_symbol(currency):
    return get_locale().get_currency_symbol(currency)


def get_decimal_symbol():
    return get_locale().get_decimal_symbol()


def get_plus_sign_symbol():
    return get_locale().get_plus_sign_symbol()


def get_minus_sign_symbol():
    return get_locale().get_minus_sign_symbol()


def get_exponential_symbol():
    return get_locale().get_exponential_symbol()


def get_group_symbol():
    return get_locale().get_group_symbol()


def format_number(number):
    return get_locale().format_number(number)


def format_decimal(number, format=None):
    return get_locale().format_decimal(number, format)


def format_currency(number, currency, format=None):
    return get_locale().format_currency(number, currency, format)


def format_percent(number, format=None):
    return get_locale().format_percent(number, format)


def format_scientific(number, format=None):
    return get_locale().format_scientific(number, format)


def parse_number(string):
    return get_locale().parse_number(string)


def parse_decimal(string):
    return get_locale().parse_decimal(string)


def to_timezone(dt):
    return get_locale().to_timezone(dt)


def to_utc(dt):
    return get_locale().to_utc(dt)


def format_time(t=None, format='medium'):
    return get_locale().format_time(t, format)


def format_date(d=None, format='medium'):
    return get_locale().format_date(d, format)


def format_datetime(dt=None, format='medium'):
    return get_locale().format_datetime(dt, format)


def parse_time(string):
    return get_locale().parse_time(string)


def parse_date(string):
    return get_locale().parse_date(string)


# -----------------------------------------------------------------------------

# Locale API
# ----------

_translations_cache = {}  # Already loaded translation objects


class DummyTranslation(object):
    """Identity translation
    """
    gettext = ugettext = _ = lazy_gettext = lazy_ugettext = _L = lambda self, msg: msg

    ngettext = ungettext = _N = lazy_ngettext = lazy_ungettext = _LN = lambda self, singular, plural, n: singular if n == 1 else plural


class Locale(CoreLocale):
    def __init__(
        self,
        language='en', territory=None, script=None, variant=None,
        dirname=None, domain=None,
        timezone=None, default_timezone=None
    ):
        """
        A locale

        In:
          - ``language`` -- the language code (i.e 'en'),
            as described in `ISO 639 <http://ftp.ics.uci.edu/pub/ietf/http/related/iso639.txt>`_
          - ``territory`` -- the territory (country or region) code (i.e 'US'),
            as described in `ISO 3166 <http://userpage.chemie.fu-berlin.de/diverse/doc/ISO_3166.html>`_
          - ``script`` -- the script code,
            as described in `ISO 15924 <http://www.evertype.com/standards/iso15924/>`_
          - ``variant`` -- the variant code (i.e ``Locale('de', 'DE', '1901')``)

          - ``dirname`` -- the directory containing the ``MO`` files
          - ``domain`` -- the messages domain

          - ``timezone`` -- the timezone (timezone object or code)
            (i.e ``pytz.timezone('America/Los_Angeles')`` or ``America/Los_Angeles``)
          - ``default_timezone`` -- default timezone when a ``datetime`` object has
            no associated timezone. If no default timezone is given, the ``timezone``
            value is used
        """
        super(Locale, self).__init__(language, territory, script, variant)

        self.domain = domain
        self.translation_directories = {}
        if dirname is not None:
            self.add_translation_directory(dirname, domain)

        if isinstance(timezone, basestring):
            timezone = pytz.timezone(timezone)
        self.tzinfo = timezone

        if default_timezone is None:
            default_timezone = timezone

        if isinstance(default_timezone, basestring):
            default_timezone = pytz.timezone(default_timezone)
        self.default_timezone = default_timezone

        self._previous_locales = []

        # By default, load the translation files for the 'nagare' domain
        # from the nagare directories
        nagare_translations_dir = pkg_resources.resource_filename(pkg_resources.Requirement.parse('nagare'), 'data/locale')
        self.add_translation_directory(nagare_translations_dir, 'nagare')

        self.zone_formats['region'] = '%s'

    def add_translation_directory(self, dirname, domain=None):
        """Associate a directory to a translation domain

        In:
          - ``dirname`` -- the directory
          - ``domain`` -- the translation domain
        """
        self.translation_directories[domain] = dirname

    def has_translation_directory(self, domain=None):
        """Test if a domain has an associated directory

        In:
          - ``domain`` -- the translation domain

        Return:
          - bool
        """
        return domain in self.translation_directories

    def get_translation_directory(self, domain=None):
        """Return the directory associated to the domain

        In:
          - ``domain`` -- the translation domain
        """
        return self.translation_directories.get(domain)

    def _get_translation(self, domain=None):
        """Load the translation object, if not already loaded

        In:
          - ``domain`` -- translation domain

        Return:
          - translation object
        """
        if self.language is None:
            return DummyTranslation()

        domain = domain or self.domain
        dirname = self.get_translation_directory(domain) or self.get_translation_directory(None)
        args = (dirname, self.language, domain)

        translation = _translations_cache.get(args)
        if translation is None:
            translation = support.Translations.load(*args)
            _translations_cache[args] = translation

        return translation

    def gettext(self, msg, domain=None, **kw):
        """Return the localized translation of a message

        In:
          - ``msg`` -- message to translate
          - ``domain`` -- translation domain
          - ``kw`` -- optional values to substitute into the translation

        Return:
          - the localized translation, as a 8-bit string encoded with the
            catalog's charset encoding
        """
        msg = self._get_translation(domain).gettext(msg)
        return msg % kw if kw else msg

    def ugettext(self, msg, domain=None, **kw):
        """Return the localized translation of a message

        In:
          - ``msg`` -- message to translate
          - ``domain`` -- translation domain
          - ``kw`` -- optional values to substitute into the translation

        Return:
          - the localized translation, as an unicode string
        """
        msg = self._get_translation(domain).ugettext(msg)
        return msg % kw if kw else msg
    _ = ugettext

    def ngettext(self, singular, plural, n, domain=None, **kw):
        """Return the plural-forms localized translation of a message

        If a translation is found, apply the ``plural`` formula to ``n``, and
        return the resulting translation. If no translation is found, return
        ``singular`` if ``n`` is 1; return ``plural`` otherwise

        In:
          - ``singular`` -- singular form of the message
          - ``plural`` -- plural form of the message
          - ``n`` -- singular or plural form wanted?
          - ``domain`` -- translation domain
          - ``kw`` -- optional values to substitute into the translation

        Return:
          - the localized translation, as a 8-bit string encoded with the
            catalog's charset encoding
        """
        msg = self._get_translation(domain).ngettext(singular, plural, n)
        return msg % kw if kw else msg

    def ungettext(self, singular, plural, n, domain=None, **kw):
        """Return the plural-forms localized translation of a message

        If a translation is found, apply the ``plural`` formula to ``n``, and
        return the resulting translation. If no translation is found, return
        ``singular`` if ``n`` is 1; return ``plural`` otherwise

        In:
          - ``singular`` -- singular form of the message
          - ``plural`` -- plural form of the message
          - ``n`` -- singular or plural form wanted?
          - ``domain`` -- translation domain
          - ``domain`` -- translation domain
          - ``kw`` -- optional values to substitute into the translation

        Return:
          - the localized translation, as an unicode string
        """
        msg = self._get_translation(domain).ungettext(singular, plural, n)
        return msg % kw if kw else msg
    _N = ungettext

    def lazy_gettext(self, msg, domain=None, **kw):
        """Return the lazy localized translation of a message

        In:
          - ``msg`` -- message to translate
          - ``domain`` -- translation domain
          - ``kw`` -- optional values to substitute into the translation

        Return:
          - the lazy localized translation, as a 8-bit string encoded with the
            catalog's charset encoding
        """
        return LazyProxy(self.gettext, msg, domain, **kw)

    def lazy_ugettext(self, msg, domain=None, **kw):
        """Return the lazy localized translation of a message

        In:
          - ``msg`` -- message to translate
          - ``domain`` -- translation domain
          - ``kw`` -- optional values to substitute into the translation

        Return:
          - the lazy localized translation, as an unicode string
        """
        return LazyProxy(self.ugettext, msg, domain, **kw)
    _L = lazy_ugettext

    def lazy_ngettext(self, singular, plural, n, domain=None, **kw):
        """Return the lazy plural-forms localized translation of a message

        If a translation is found, apply the ``plural`` formula to ``n``, and
        return the resulting translation. If no translation is found, return
        ``singular`` if ``n`` is 1; return ``plural`` otherwise

        In:
          - ``singular`` -- singular form of the message
          - ``plural`` -- plural form of the message
          - ``n`` -- singular or plural form wanted?
          - ``domain`` -- translation domain
          - ``kw`` -- optional values to substitute into the translation

        Return:
          - the lazy localized translation, as a 8-bit string encoded with the
            catalog's charset encoding
        """
        return LazyProxy(self.ngettext, singular, plural, n, domain, **kw)

    def lazy_ungettext(self, singular, plural, n, domain=None, **kw):
        """Return the lazy plural-forms localized translation of a message

        If a translation is found, apply the ``plural`` formula to ``n``, and
        return the resulting translation. If no translation is found, return
        ``singular`` if ``n`` is 1; return ``plural`` otherwise

        In:
          - ``singular`` -- singular form of the message
          - ``plural`` -- plural form of the message
          - ``n`` -- singular or plural form wanted?
          - ``domain`` -- translation domain
          - ``kw`` -- optional values to substitute into the translation

        Return:
          - the lazy localized translation, as an unicode string
        """
        return LazyProxy(self.ungettext, singular, plural, n, domain, **kw)
    _LN = lazy_ungettext

    def get_period_names(self):
        """Return the names for day periods (AM/PM)

        >>> Locale('en', 'US').get_period_names()['am']
        u'AM'
        """
        return dates.get_period_names(locale=self)

    def get_day_names(self, width='wide', context='format'):
        """Return the day names for the specified format

        >>> Locale('en', 'US').get_day_names('wide')[1]
        u'Tuesday'
        >>> Locale('es').get_day_names('abbreviated')[1]
        u'mar'
        >>> Locale('de', 'DE').get_day_names('narrow', context='stand-alone')[1]
        u'D'

        In:
          - ``width`` -- 'wide', 'abbreviated' or 'narrow'
          - ``context`` -- either 'format' or 'stand-alone'

        Return:
          - the day names
        """
        return dates.get_day_names(width, context, self)

    def get_month_names(self, width='wide', context='format'):
        """Return the month names for the specified format

        >>> Locale('en', 'US').get_month_names('wide')[1]
        u'January'
        >>> Locale('es').get_month_names('abbreviated')[1]
        u'ene'
        >>> Locale('de', 'DE').get_month_names('narrow', context='stand-alone')[1]
        u'J'

        In:
          - ``width`` -- 'wide', 'abbreviated' or 'narrow'
          - ``context`` -- either 'format' or 'stand-alone'

        Return:
          - the month names
        """
        return dates.get_month_names(width, context, self)

    def get_quarter_names(self, width='wide', context='format'):
        """Return the quarter names for the specified format

        >>> Locale('en', 'US').get_quarter_names('wide')[1]
        u'1st quarter'
        >>> Locale('de', 'DE').get_quarter_names('abbreviated')[1]
        u'Q1'

        In:
          - ``width`` -- 'wide', 'abbreviated' or 'narrow'
          - ``context`` -- either 'format' or 'stand-alone'

        Return:
          - the quarter names
        """
        return dates.get_quarter_names(width, context, self)

    def get_era_names(self, width='wide'):
        """Return the era names used for the specified format

        >>> Locale('en', 'US').get_era_names('wide')[1]
        u'Anno Domini'
        >>> Locale('de', 'DE').get_era_names('abbreviated')[1]
        u'n. Chr.'

        In:
          - ``width`` -- 'wide', 'abbreviated' or 'narrow'

        Return:
          - the era names
        """
        return dates.get_era_names(width, self)

    def get_date_format(self, format='medium'):
        """Return the date formatting pattern for the specified format

        >>> Locale('en', 'US').get_date_format()
        <DateTimePattern u'MMM d, yyyy'>
        >>> Locale('de', 'DE').get_date_format('full')
        <DateTimePattern u'EEEE, d. MMMM yyyy'>

        In:
          - ``format`` -- 'full', 'long', 'medium' or 'short'

        Return:
          - the date formatting pattern
        """
        return dates.get_date_format(format, self)

    def get_datetime_format(self, format='medium'):
        """Return the datetime formatting pattern for the specified format

        >>> Locale('en', 'US').get_datetime_format()
        u'{1} {0}'

        In:
          - ``format`` -- 'full', 'long', 'medium' or 'short'

        Return:
          - the datetime formatting pattern
        """
        return dates.get_datetime_format(format, self)

    def get_time_format(self, format='medium'):
        """Return the time formatting pattern for the specified format

        >>> Locale('en', 'US').get_time_format()
        <DateTimePattern u'h:mm:ss a'>
        >>> Locale('de', 'DE').get_time_format('full')
        <DateTimePattern u'HH:mm:ss v'>

        In:
          - ``format`` -- 'full', 'long', 'medium' or 'short'

        Return:
          - the time formatting pattern
        """
        return dates.get_time_format(format, self)

    def get_timezone_gmt(self, dt=None, width='long'):
        """Return the timezone associated with the given ``dt`` datetime object, formatted as string indicating the offset from GMT

        >>> dt = datetime.datetime(2007, 4, 1, 15, 30)
        >>> Locale('en').get_timezone_gmt(dt)
        u'GMT+00:00'

        In:
          - ``datetime`` -- a ``datetime`` object; if ``None``, the current
            date and time in UTC is used
          - ``width`` -- either 'long' or 'short'

        Return:
          - The timezone
        """
        return dates.get_timezone_gmt(dt, width, self)

    def get_timezone_location(self, dt_or_timezone=None):
        """Return a representation of the given timezone using "location format"

        The result depends on both the local display name of the country and the
        city associated with the time zone:

        >>> from pytz import timezone
        >>> tz = timezone('America/St_Johns')
        >>> Locale('de', 'DE').get_timezone_location(tz)
        u"Kanada (St. John's)"
        >>> tz = timezone('America/Mexico_City')
        >>> Locale('de', 'DE').get_timezone_location(tz)
        u'Mexiko (Mexiko-Stadt)'

        In:
          - ``dt_or_tzinfo`` -- ``datetime`` or ``tzinfo`` object that determines
            the timezone; if `None`, the current date and time in UTC is assumed

        Return:
          - the timezone representation
        """
        return dates.get_timezone_location(dt_or_timezone, self)

    def get_timezone_name(self, dt_or_timezone=None, width='long', uncommon=False):
        """Return the localized display name for the given timezone

        In:
          - ``dt_or_timezone`` -- the timezone, specified using a ``datetime`` or ``tzinfo`` object
          - ``width`` --
          - ``uncommon`` --

        >>> from pytz import timezone
        >>> dt = datetime.time(15, 30, tzinfo=timezone('America/Los_Angeles'))
        >>> Locale('en', 'US').get_timezone_name(dt)
        u'Pacific Standard Time'
        >>> Locale('en', 'US').get_timezone_name(dt, width='short')
        u'PST'

        In:
          - ``dt_or_tzinfo`` -- the ``datetime`` or ``tzinfo`` object that determines
            the timezone; if a ``tzinfo`` object is used, the resulting display
            name will be generic, i.e. independent of daylight savings time;
            if ``None``, the current date in UTC is assumed
          - ``width`` -- either 'long' or 'short'
          - ``uncommon`` -- whether even uncommon timezone abbreviations should be used

        Return:
          - the localized timezone name
        """
        return dates.get_timezone_name(dt_or_timezone, width, uncommon, self)

    def get_currency_name(self, currency, count=None):
        """Return the name used for the specified currency

        >>> Locale('en', 'US').get_currency_name('USD')
        u'US Dollar'

        In:
          - ``currency`` -- the currency code
          - ``count`` -- If provided the currency name will be pluealized to that number

        Return:
          - the currency name
        """
        return numbers.get_currency_name(currency, count, self)

    def get_currency_symbol(self, currency):
        """Return the symbol used for the specified currency

        >>> Locale('en', 'US').get_currency_symbol('USD')
        u'$'

        In:
          - ``currency`` -- the currency code

        Return:
          - the currency symbol
        """
        return numbers.get_currency_symbol(currency, self)

    def get_decimal_symbol(self):
        """Return the symbol used to separate decimal fractions

        >>> Locale('en', 'US').get_decimal_symbol()
        u'.'
        """
        return numbers.get_decimal_symbol(self)

    def get_plus_sign_symbol(self):
        """Return the plus sign symbol

        >>> Locale('en', 'US').get_plus_sign_symbol()
        u'+'
        """
        return numbers.get_plus_sign_symbol(self)

    def get_minus_sign_symbol(self):
        """Return the plus sign symbol

        >>> Locale('en', 'US').get_minus_sign_symbol()
        u'-'
        """
        return numbers.get_minus_sign_symbol(self)

    def get_exponential_symbol(self):
        """Return the symbol used to separate mantissa and exponent

        >>> Locale('en', 'US').get_exponential_symbol()
        u'E'
        """
        return numbers.get_exponential_symbol(self)

    def get_group_symbol(self):
        """Return the symbol used to separate groups of thousands

        >>> Locale('en', 'US').get_group_symbol()
        u','
        """
        return numbers.get_group_symbol(self)

    def format_number(self, number):
        """Return the given number formatted

        >>> Locale('en', 'US').format_number(1099)
        u'1,099'
        """
        return numbers.format_number(number, self)

    def format_decimal(self, number, format=None):
        """Return the given decimal number formatted

        >>> Locale('en', 'US').format_decimal(1.2345)
        u'1.234'
        >>> Locale('en', 'US').format_decimal(1.2346)
        u'1.235'
        >>> Locale('en', 'US').format_decimal(-1.2346)
        u'-1.235'
        >>> Locale('sv', 'SE').format_decimal(1.2345)
        u'1,234'
        >>> Locale('de').format_decimal(12345)
        u'12.345'
        """
        return numbers.format_decimal(number, format, self)

    def format_currency(self, number, currency, format=None):
        """Return formatted currency value

        >>> Locale('en', 'US').format_currency(1099.98, 'USD')
        u'$1,099.98'
        >>> Locale('es', 'CO').format_currency(1099.98, 'USD')
        u'US$\\xa01.099,98'
        >>> Locale('de', 'DE').format_currency(1099.98, 'EUR')
        u'1.099,98\\xa0\\u20ac'

        The pattern can also be specified explicitly:

        >>> Locale('en', 'US').format_currency(1099.98, 'EUR', u'\xa4\xa4 #,##0.00')
        u'EUR 1,099.98'
        """
        return numbers.format_currency(number, currency, format, self)

    def format_percent(self, number, format=None):
        """Return formatted percent value

        >>> Locale('en', 'US').format_percent(0.34)
        u'34%'
        >>> Locale('en', 'US').format_percent(25.1234)
        u'2,512%'
        >>> Locale('sv', 'SE').format_percent(25.1234)
        u'2\\xa0512\\xa0%'

        The format pattern can also be specified explicitly:

        >>> Locale('en', 'US').format_percent(25.1234, u'#,##0\u2030')
        u'25,123\u2030'
        """
        return numbers.format_percent(number, format, self)

    def format_scientific(self, number, format=None):
        """Return value formatted in scientific notation

        >>> Locale('en', 'US').format_scientific(10000)
        u'1E4'

        The format pattern can also be specified explicitly:

        >>> Locale('en', 'US').format_scientific(1234567, u'##0E00')
        u'1.23E06'
        """
        return numbers.format_scientific(number, format, self)

    def parse_number(self, string):
        """Parse localized number string into a long integer

        >>> Locale('en', 'US').parse_number('1,099')
        1099L
        >>> Locale('de', 'DE').parse_number('1.099')
        1099L

        When the given string cannot be parsed, an exception is raised:

        >>> Locale('de').parse_number('1.099,98')
        Traceback (most recent call last):
            ...
        NumberFormatError: '1.099,98' is not a valid number
        """
        return numbers.parse_number(string, self)

    def parse_decimal(self, string):
        """Parse localized decimal string into a float

        >>> Locale('en', 'US').parse_decimal('1,099.98')
        1099.98
        >>> Locale('de').parse_decimal('1.099,98')
        1099.98

        When the given string cannot be parsed, an exception is raised:

        >>> Locale('de').parse_decimal('2,109,998')
        Traceback (most recent call last):
            ...
        NumberFormatError: '2,109,998' is not a valid decimal number
        """
        return numbers.parse_decimal(string, self)

    def to_timezone(self, dt):
        """Return a localized datetime object

        In:
          - ``dt`` -- ``datetime`` object

        Return:
          - new localized ``datetime`` object
        """
        if not self.tzinfo:
            return dt

        if not dt.tzinfo:
            dt = (self.default_timezone or pytz.UTC).localize(dt)

        return dt.astimezone(self.tzinfo)

    def to_utc(self, dt):
        """Return a UTC datetime object

        In:
          - ``dt`` -- ``datetime`` object

        Return:
          - new localized to UTC ``datetime`` object
        """
        if not dt.tzinfo:
            dt = (self.default_timezone or pytz.UTC).localize(dt)

        return dt.astimezone(pytz.UTC)

    def format_time(self, t=None, format='medium'):
        """Return a time formatted according to the given pattern

        >>> t = datetime.time(15, 30)
        >>> Locale('en', 'US').format_time(t)
        u'3:30:00 PM'
        >>> Locale('de', 'DE').format_time(t, format='short')
        u'15:30'

        If you don't want to use the locale default formats, you can specify a
        custom time pattern:

        >>> Locale('en').format_time(t, "hh 'o''clock' a")
        u"03 o'clock PM"

        In:
          - ``t`` --  ``time`` or ``datetime`` object; if `None`, the current time in UTC is used
          - ``format`` -- 'full', 'long', 'medium', or 'short', or a custom date/time pattern

        Returns:
          - the formatted time string
        """
        if isinstance(t, datetime.time):
            d = datetime.datetime.now()
            t = d.replace(hour=t.hour, minute=t.minute, second=t.second)

        if isinstance(t, datetime.datetime):
            t = self.to_utc(t)

        return dates.format_time(t, format, locale=self, tzinfo=self.tzinfo)

    def format_date(self, d=None, format='medium'):
        """Return a date formatted according to the given pattern

        >>> d = datetime.date(2007, 04, 01)
        >>> Locale('en', 'US').format_date(d)
        u'Apr 1, 2007'
        >>> Locale('de', 'DE').format_date(d, format='full')
        u'Sonntag, 1. April 2007'

        If you don't want to use the locale default formats, you can specify a
        custom date pattern:

        >>> Locale('en').format_date(d, "EEE, MMM d, ''yy")
        u"Sun, Apr 1, '07"

        In:
          - ``d`` -- ``date`` or ``datetime`` object; if ``None``, the current date is used
          - ``format`` -- 'full', 'long', 'medium', or 'short', or a custom date/time pattern

        Return:
          - the formatted date string
        """
        return dates.format_date(d, format, self)

    def format_datetime(self, dt=None, format='medium'):
        """Return a date formatted according to the given pattern

        >>> dt = datetime.datetime(2007, 04, 01, 15, 30)
        >>> Locale('en', 'US').format_datetime(dt)
        u'Apr 1, 2007 3:30:00 PM'

        >>> from pytz import timezone
        >>> dt = datetime.datetime(2007, 04, 01, 15, 30)
        >>> Locale('fr', 'FR', timezone=timezone('Europe/Paris'), default_timezone=pytz.UTC).format_datetime(dt, 'full')
        u'dimanche 1 avril 2007 17:30:00 HEC'
        >>> Locale('en', timezone=timezone('US/Eastern'), default_timezone=pytz.UTC).format_datetime(dt, "yyyy.MM.dd G 'at' HH:mm:ss zzz")
        u'2007.04.01 AD at 11:30:00 EDT'

        In:
          - ``dt`` -- ``datetime`` object; if ``None``, the current date and time is used
          - ``format`` -- 'full', 'long', 'medium', or 'short', or a custom date/time pattern

        Return:
          - The formatted datetime string
        """
        if dt:
            dt = self.to_timezone(dt)
        return dates.format_datetime(dt, format, locale=self, tzinfo=self.tzinfo)

    def parse_time(self, string):
        """Parse a time from a string

        This function uses the time format for the locale as a hint to determine
        the order in which the time fields appear in the string.

        >>> Locale('en', 'US').parse_time('15:30:00')
        datetime.time(15, 30)

        In:
          - ``string`` -- the string containing the time

        Return:
          - a ``datetime.time`` object
        """
        return dates.parse_time(string, self)

    def parse_date(self, string):
        """Parse a date from a string

        This function uses the date format for the locale as a hint to determine
        the order in which the date fields appear in the string.

        >>> Locale('en', 'US').parse_date('4/1/04')
        datetime.date(2004, 4, 1)
        >>> Locale('de', 'DE').parse_date('01.04.2004')
        datetime.date(2004, 4, 1)

        In:
          - ``string`` -- the string containing the date

        Return:
          - a ``datetime.datetime`` object
        """
        return dates.parse_date(string, self)

    def __enter__(self):
        """Push this locale to the stack
        """
        previous_locale = get_locale()

        if not self.translation_directories:
            self.translation_directories.update(previous_locale.translation_directories)

        self._previous_locales.append(previous_locale)
        set_locale(self)

    def __exit__(self, *args, **kw):
        """Pop this locale from the stack
        """
        set_locale(self._previous_locales.pop())


# -----------------------------------------------------------------------------

class NegotiatedLocale(Locale):
    def __init__(
        self,
        request,
        locales, default_locale=(None, None),
        dirname=None, domain=None,
        timezone=None, default_timezone=None
    ):
        """
        A locale with negotiated language and territory

        In:
          - ``request`` -- the HTTP request object
          - ``locales`` -- tuples of (language, territory) accepted by the application
            (i.e ``[('fr', 'FR'), ('de', 'DE'), ('en',)]``)
          - ``default_locale`` -- tuple of (language, territory) to use if the
            negociation failed

          - ``dirname`` -- the directory containing the ``MO`` files
          - ``domain`` -- the messages domain

          - ``timezone`` -- the timezone (timezone object or code)
            (i.e ``pytz.timezone('America/Los_Angeles')`` or ``America/Los_Angeles``)
          - ``default_timezone`` -- default timezone when a ``datetime`` object has
            no associated timezone. If no default timezone is given, the ``timezone``
            value is used
        """
        locale = negotiate_locale(
            request.accept_language,
            map('-'.join, locales),
            '-'
        )

        if not locale:
            (language, territory) = (default_locale + (None,))[:2]
        else:
            locale = core.LOCALE_ALIASES.get(locale, locale).replace('_', '-')

            if '-' not in locale:
                language = locale
                territory = None
            else:
                (language, territory) = locale.split('-')
                territory = territory.upper()

        super(NegotiatedLocale, self).__init__(
            language, territory,
            dirname=dirname, domain=domain,
            timezone=timezone, default_timezone=default_timezone
        )


# -----------------------------------------------------------------------------

def get_locale():
    return getattr(local.request, 'locale', None) or Locale()


def set_locale(locale):
    local.request.locale = locale



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --


class ExpirationError(LookupError):
    """Raised when a session or a state id is no longer valid
    """
    pass


class SessionSecurityError(LookupError):
    """Raised when the secure id of a session is not valid
    """
    pass



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

from nagare import component, presentation, continuation, var
from nagare.namespaces import xhtml


class Foo(object):
    def __init__(self):
        self.my_property = "I'm foo"

    def set_my_property(self, v):
        self.my_property = v


class Bar(Foo):
    def __init__(self):
        self.set_my_property("I'm bar")


class Foobar:
    def __init__(self):
        self.my_property = component.Component(Foo())


# -------------------------------------------------------------------------------------------------------

class App1:
    def __init__(self):
        self.foo = Foo()

        self.bar = component.Component(Bar())
        self.bar.on_answer(self.foo.set_my_property)


def test1():
    """Component - test on_answer/answer"""
    app = App1()

    assert app.foo.my_property == "I'm foo"

    app.bar.answer("I'm bar")
    assert app.foo.my_property == "I'm bar"

    app.bar.answer("I'm foo again")
    assert app.foo.my_property == "I'm foo again"


# -------------------------------------------------------------------------------------------------------

def test2():
    """Component - becomes"""
    app = Foobar()

    assert isinstance(app.my_property(), Foo)
    assert not bool(app.my_property.model)
    assert app.my_property().my_property == "I'm foo"

    app.my_property.becomes(Bar(), model='bar')

    assert isinstance(app.my_property(), Bar)
    assert app.my_property.model == 'bar'
    assert app.my_property().my_property == "I'm bar"

    app.my_property.becomes(Foo())
    assert isinstance(app.my_property(), Foo)
    assert not bool(app.my_property.model)
    assert app.my_property().my_property == "I'm foo"


# -------------------------------------------------------------------------------------------------------

if continuation.has_continuation:
    def test3():
        """Component - call"""
        v = var.Var(42)
        app = Foobar()

        app.my_property.becomes(Foo(), model='foo')
        assert isinstance(app.my_property(), Foo)
        assert app.my_property.model == 'foo'

        continuation.call_wrapper(lambda: v(app.my_property.call(Bar(), model='bar')))

        assert isinstance(app.my_property(), Bar)
        assert app.my_property.model == 'bar'

        app.my_property.answer("I'm bar")

        assert isinstance(app.my_property(), Foo)
        assert app.my_property.model == 'foo'
        assert v() == "I'm bar"


# -------------------------------------------------------------------------------------------------------

@presentation.render_for(Bar)
def render(self, h, *args):
    return h.table(h.tr(h.td("I'm Bar")))


@presentation.render_for(Bar, model='foo')
def render_bar(self, h, *args):
    return h.h1("I'm bar in foo")


def test4():
    """Component - render"""
    foo = component.Component(Foo())

    h = xhtml.Renderer()
    try:
        # No default model for Foo() objects
        foo.render(h)
    except:
        pass
    else:
        assert False

    foo.becomes(Bar())
    assert foo.render(h).write_htmlstring(pretty_print=True).strip() == "<table><tr><td>I'm Bar</td></tr></table>"

    foo.becomes(model='foo')
    assert foo.render(h).write_htmlstring(pretty_print=True).strip() == "<h1>I'm bar in foo</h1>"



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Base classes for the sessions management"""

import random

import configobj

from nagare import config
from nagare.admin import reference
from nagare.sessions import SessionSecurityError, serializer


class State(object):
    """A state (objects graph serialized / de-serialized by a sessions manager)
    """
    def __init__(self, sessions_manager, session_id, state_id, secure_id, use_same_state):
        """Initialization

        In:
          - ``sessions_manager`` -- the manager of this state
          - ``session_id`` -- session id of this state
          - ``state_id`` -- id of this state (``None`` to create a new state)
          - ``secure_id`` -- the secure number associated to the session
          - ``use_same_state`` -- is a copy of this state to create?
        """
        self.sessions_manager = sessions_manager
        self.session_id = session_id
        self.state_id = state_id
        self.secure_id = secure_id
        self.use_same_state = use_same_state

        self.back_used = False  # Is this state a snapshot of a previous objects graph?
        self.lock = (sessions_manager.create_lock if state_id is None else sessions_manager.get_lock)(self.session_id)

    def sessionid_in_url(self, request, response):
        """Return the session and states ids to put into an URL

        In:
          - ``request`` -- the web request
          - ``response`` -- the web response

        Return:
          - session id parameter
          - state id parameter
        """
        return self.sessions_manager.sessionid_in_url(self.session_id, self.state_id, request, response)

    def sessionid_in_form(self, h, request, response):
        """Return the DOM tree to merge into a form, to add the session and state hidden ids

        In:
          - ``h`` -- the current renderer
          - ``request`` -- the web request
          - ``response`` -- the web response

        Return:
          - the DOM tree
        """
        return self.sessions_manager.sessionid_in_form(self.session_id, self.state_id, h, request, response)

    def acquire(self):
        """Lock the state
        """
        self.lock.acquire()  # Lock the session

    def release(self):
        """Release the state
        """
        self.lock.release()  # Release the session

    def get_root(self):
        """Retrieve the objects graph of this state

        Return:
         - the objects graph (``None`` if this state is new)
        """
        if self.state_id is None:
            # New state
            self.sessions_manager.create(self.session_id, self.secure_id, self.lock)
            self.state_id = 0
            data = None
        else:
            # Existing state
            new_state_id, secure_id, data = self.sessions_manager.get_root(self.session_id, self.state_id)
            if secure_id != self.secure_id:
                raise SessionSecurityError

            if not self.use_same_state:
                self.back_used = (self.state_id != new_state_id - 1)
                self.state_id = new_state_id

        return data

    def set_root(self, use_same_state, data):
        """Store the objects graph of this state

        In:
          - ``use_same_state`` -- is the objects graph to be stored in this state or in a new one?
          - ``data`` -- the objects graph
        """
        self.sessions_manager.set_root(self.session_id, self.state_id, self.secure_id, self.use_same_state or use_same_state, data)

    def delete(self):
        """Delete the session of this state
        """
        self.sessions_manager.delete(self.session_id)


class Sessions(object):
    """The sessions managers
    """
    spec = {
        'security_cookie_httponly': 'boolean(default=True)',
        'security_cookie_name': 'string(default="_nagare")',
        'security_cookie_secure': 'boolean(default=False)',
        'states_history': 'boolean(default=True)',
        'pickler': 'string(default="cPickle:Pickler")',
        'unpickler': 'string(default="cPickle:Unpickler")',
        'serializer': 'string(default="nagare.sessions.serializer:Dummy")'
    }

    def __init__(
        self,
        states_history=True,
        security_cookie_httponly=True,
        security_cookie_name='_nagare',
        security_cookie_secure=False,
        serializer=serializer.Dummy, pickler=None, unpickler=None
    ):
        """Initialization

        In:
          - ``states_history`` -- are all the states kept or only the latest?
          - ``security_cookie_name`` -- name of the cookie where the session secure id is stored
          - ``serializer`` -- serializer / deserializer of the states
          - ``pickler`` -- pickler used by the serializer
          - ``unpickler`` -- unpickler used by the serializer
        """
        self.states_history = states_history
        self.security_cookie_httponly = security_cookie_httponly
        self.security_cookie_name = security_cookie_name
        self.security_cookie_secure = security_cookie_secure
        self.serializer = serializer(pickler, unpickler)

    def set_config(self, filename, conf, error):
        """Read the configuration parameters

        In:
          - ``filename`` -- the path to the configuration file
          - ``conf`` -- the ``ConfigObj`` object, created from the configuration file
          - ``error`` -- the function to call in case of configuration errors
        """
        conf = dict([(k, v) for k, v in conf.items() if k in self.spec])
        conf = configobj.ConfigObj(conf, configspec=self.spec)
        config.validate(filename, conf, error)

        self.states_history = conf['states_history']
        self.security_cookie_name = conf['security_cookie_name']

        pickler = reference.load_object(conf['pickler'])[0]
        unpickler = reference.load_object(conf['unpickler'])[0]
        serializer = reference.load_object(conf['serializer'])[0]
        self.serializer = serializer(pickler, unpickler)

        return conf

    def sessionid_in_url(self, session_id, state_id, request, response):
        """Return the session and states ids to put into an URL

        In:
          - ``request`` -- the web request
          - ``response`` -- the web response

        Return:
          - session id parameter
          - state id parameter (optional, only if ``states_history`` is ``True``)
        """
        ids = ('_s=%d' % session_id,)
        if self.states_history:
            ids += ('_c=%05d' % state_id,)

        return ids

    def sessionid_in_form(self, session_id, state_id, h, request, response):
        """Return the DOM tree to merge into a form to add the session and state
        hidden ids

        In:
          - ``h`` -- the current renderer
          - ``request`` -- the web request object
          - ``response`` -- the web response object

        Return:
          - a DOM tree
        """
        return (
            h.input(name='_s', value=session_id, type='hidden'),
            h.input(name='_c', value='%05d' % state_id, type='hidden')
        )

    def _get_ids(self, request):
        """Search the session id and the state id into the request parameters

        In:
          - ``request`` -- the web request

        Return:
          - session id
          - state id
        """
        return (
            int(request.params['_s']),
            int(request.params['_c']) if self.states_history else 0
        )

    def check_session_id(self, session_id):
        """Test if a session exist

        In:
          - ``session_id`` -- id of a session

        Return:
          - is ``session_id`` the id of an existing session?
        """
        return False

    def get_state(self, request, response, use_same_state):
        """Create a new state or return an existing one

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object
          - ``use_same_state`` -- is a copy of the state to created?

        Return:
          - the state
        """
        try:
            session_id, state_id = self._get_ids(request)
        except (KeyError, ValueError, TypeError):
            state_id = None

            # Create a new session id
            while True:
                session_id = random.randint(1000000000000000, 9999999999999999)
                if not self.check_session_id(session_id):
                    break

        secure_id = None
        if self.security_cookie_name:
            secure_id = request.cookies.get(self.security_cookie_name)
            if not secure_id:
                secure_id = str(random.randint(1000000000000000, 9999999999999999))
                response.set_cookie(self.security_cookie_name, secure_id, path=request.script_name + '/',
                                    secure=self.security_cookie_secure,
                                    httponly=self.security_cookie_httponly)

        return State(self, session_id, state_id, secure_id, use_same_state or not self.states_history)

    def get_root(self, session_id, state_id):
        """Retrieve the objects graph of a state

        In:
          - ``session_id`` -- session id of this state
          - ``state_id`` -- id of this state

        Return:
          - id of the latest state
          - secure number associated to the session
          - objects graph
        """
        new_state_id, secure_id, session_data, state_data = self.fetch_state(session_id, state_id)
        return new_state_id, secure_id, self.serializer.loads(session_data, state_data)

    def set_root(self, session_id, state_id, secure_id, use_same_state, data):
        """Store the state

        In:
          - ``session_id`` -- session id of this state
          - ``state_id`` -- id of this state
          - ``secure_id`` -- the secure number associated to the session
          - ``use_same_state`` -- is a copy of this state to be created?
          - ``data`` -- the objects graph
        """
        session_data, state_data = self.serializer.dumps(data, not use_same_state)
        self.store_state(session_id, state_id, secure_id, use_same_state, session_data, state_data)

    # -------------------------------------------------------------------------

    def create_lock(self, session_id):
        """Create a new lock for a session

        In:
          - ``session_id`` -- session id

        Return:
          - the lock
        """
        return self.get_lock(session_id)

    def get_lock(self, session_id):
        """Retrieve the lock of a session

        In:
          - ``session_id`` -- session id

        Return:
          - the lock
        """
        raise NotImplementedError()

    def create(self, session_id, secure_id, lock):
        """Create a new session

        In:
          - ``session_id`` -- id of the session
          - ``secure_id`` -- the secure number associated to the session
          - ``lock`` -- the lock of the session
        """
        raise NotImplementedError()

    def delete(self, session_id):
        """Delete a session

        In:
          - ``session_id`` -- id of the session to delete
        """
        raise NotImplementedError()

    def fetch_state(self, session_id, state_id):
        """Retrieve a state with its associated objects graph

        In:
          - ``session_id`` -- session id of this state
          - ``state_id`` -- id of this state

        Return:
          - id of the more recent stored state
          - secure number associated to the session
          - data kept into the session
          - data kept into the state
        """
        raise NotImplementedError()

    def store_state(self, session_id, state_id, secure_id, use_same_state, session_data, state_data):
        """Store a state and its associated objects graph

        In:
          - ``session_id`` -- session id of this state
          - ``state_id`` -- id of this state
          - ``secure_id`` -- the secure number associated to the session
          - ``use_same_state`` -- is this state to be stored in the previous snapshot?
          - ``session_data`` -- data to keep into the session
          - ``state_data`` -- data to keep into the state
        """
        raise NotImplementedError()



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Comet-style channels i.e HTTP push channels

This implementation is only working with a multi-threaded publisher
"""

from __future__ import with_statement

import threading
import select

from nagare import presentation, ajax


class Client(object):
    """A connected client"""

    def __init__(self, fileno, response):
        """Initialization

        In:
          - ``fileno`` -- the I/O file handle
          - ``response`` -- ``webob`` response object
        """
        self._fileno = fileno
        self._response = response

        self._event = threading.Event()

    def block(self):
        self._event.wait()

    def release(self):
        self._event.set()

    def is_blocked(self):
        return self._event.is_set()

    def fileno(self):
        return self._fileno

    def send(self, msg=None, status=None):
        if status is not None:
            self._response.status = status

        if msg is not None:
            self._response.body = msg

        if not self.is_blocked():
            self.release()


class Channel(object):
    """A comet-style channel i.e XHR long polling channel"""

    def __init__(self, id, js, history_size=0, poll_time=1000):
        """Initialization

        In:
          - ``id`` -- the channel identifier
          - ``js`` -- the javascript function to call when a message is received
          - ``history_size`` -- number of messages kept on the server
          - ``poll_time`` -- time the brower waits, before to retry to reconnect, after an error occurs
        """
        self.id = id
        self.js = js
        self.poll_time = poll_time

        self.history_size = history_size
        self.history_nb = 0
        self.history = []

        self.lock = threading.Lock()
        self.clients = set()

    def connect(self, nb, fileno, response):
        """A browser wants to be connected to this channel

        In:
          - ``nb`` -- identifier of the next wanted message
          - ``fileno`` -- the I/O file handle
          - ``response`` -- ``webob`` response object
        """
        client = Client(fileno, response)

        with self.lock:
            self.discard_disconnected_clients()

            # Is a msg already available to send ?
            (new_nb, msg) = self.get_old_msg(nb)
            if not msg:
                self.clients.add(client)

        if msg:
            # Msg found into the history
            # Send it directly
            self._send(client, new_nb, msg)
        else:
            # No msg found
            # Wait for a new msg to be sent
            client.block()

    def discard_disconnected_clients(self):
        """Forgot about the disconnected clients
        """
        if self.clients:
            clients_to_discard = select.select(self.clients, [], [], 0)[0]

            for client in clients_to_discard:
                client.release()

            self.clients = self.clients - set(clients_to_discard)

    def send(self, msg):
        """Send a message to all the connected clients

        In:
          - ``msg`` -- message to send
        """
        with self.lock:
            # Keep the msg into the history
            self.history.append(msg)

            if len(self.history) > self.history_size:
                self.history.pop(0)
                self.history_nb += 1

            clients, self.clients = self.clients, set()

        # Send the msg to all the waiting clients
        for client in clients:
            self._send(client, self.history_nb + len(self.history), msg)

    def close(self):
        """Close a channel
        """
        with self.lock:
            clients, self.clients = self.clients, set()

        # Send a "Service Unavailable" status to all the connected clients
        for client in clients:
            client.send(status=503)   # "Service Unavailable"

    def _send(self, client, nb, msg):
        """Build the message to send

        In:
          - ``client`` -- client to send the message to
          - ``nb`` -- message id
          - ``msg`` -- the message
        """
        # The raw message sent contains the msg identifier and the data
        client.send('%09d%s' % (nb, msg.encode('utf-8')))

    def get_old_msg(self, nb):
        """Check if a message in the history is ready to be sent

        In:
          - ``nb`` -- identifier of the next message wanted by the client

        Return:
          - the tuple:
            - identifier of the next msg to ask for
            - the msg if available or ``None``
        """
        if nb >= self.history_nb + len(self.history):
            return None, None

        if nb < self.history_nb:
            nb = self.history_nb

        # Return only one message
        return nb + 1, self.history[nb - self.history_nb]

    def __len__(self):
        return len(self.clients)


class TextChannel(Channel):
    def get_old_msg(self, nb):
        """Check if a message in the history is ready to be sent

        In:
          - ``nb`` -- identifier of the next message wanted by the client

        Return:
          - the tuple:
            - identifier of the next msg to ask for
            - the msg if available or ``None``
        """
        # Concatenate all the available msgs
        msgs = self.history[max(nb, self.history_nb) - self.history_nb:]
        return len(self.history) + self.history_nb, ''.join(msgs) if msgs else None


# ----------------------------------------------------------------------------

@presentation.render_for(Channel)
def render(self, h, *args):
    ajax.javascript_dependencies(h)
    h.head.javascript_url('/static/nagare/comet.js')

    return h.script('setTimeout(function () { comet_getAndEval("%s", 0, %s, %d) }, 300)' % (self.id, self.js, self.poll_time))


# ----------------------------------------------------------------------------

class Channels(dict):
    """Channels manager
    """
    channel_factory = Channel

    def create(self, id, *args, **kw):
        """Create a new channel

        In:
          - ``id`` -- the channel identifier
        """
        if id not in self:
            self[id] = self.channel_factory(id, *args, **kw)

    def connect(self, id, nb, fileno, response):
        """A browser wants to be connected to a channel

        In:
          - ``id`` -- the channel identifier
          - ``nb`` -- identifier of the next message wanted
          - ``fileno`` -- the I/O file handle
          - ``response`` -- ``webob`` response object
        """
        channel = self.get(id)

        if channel is None:
            response.status = 404  # "Not Found"
        else:
            self.discard_disconnected_clients()

            channel.connect(nb, fileno, response)

    def discard_disconnected_clients(self):
        for channel in self.values():
            channel.discard_disconnected_clients()

    def send(self, id, msg):
        """Send a message to all the connected clients of a channel

        In:
          - ``id`` -- the channel identifier
          - ``msg`` -- message to send
        """
        self[id].send(msg)

    def has_clients(self, id):
        self.discard_disconnected_clients()
        return len(self[id])

    def __delitem__(self, id):
        """Close a channel

        In:
          - ``id`` -- the channel identifier
        """
        channel = self.pop(id)
        channel.close()


class TextChannels(Channels):
    channel_factory = TextChannel


channels = TextChannels()



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

import StringIO

import logging
import logging.config

import configobj

from nagare import local


# -----------------------------------------------------------------------------

logging._srcfile = __file__[:-1] if __file__.endswith(('.pyc', '.pyo')) else __file__


# -----------------------------------------------------------------------------

def get_logger(name='.'):
    if name is None:
        name = '.'

    if name.startswith('.'):
        try:
            name = (local.request.logger_name + name).rstrip('.')
        except AttributeError:
            name = 'nagare.application'

    return logging.getLogger(name)


def set_logger(name):
    local.request.logger_name = name


def debug(msg, *args, **kw):
    get_logger().debug(msg, *args, **kw)


def info(msg, *args, **kw):
    get_logger().info(msg, *args, **kw)


def warning(msg, *args, **kw):
    get_logger().warning(msg, *args, **kw)


def error(msg, *args, **kw):
    get_logger().error(msg, *args, **kw)


def critical(msg, *args, **kw):
    get_logger().critical(msg, *args, **kw)


def exception(msg, *args):
    get_logger().exception(msg, *args)


def log(level, msg, *args, **kw):
    get_logger().log(level, msg, *args, **kw)


# -----------------------------------------------------------------------------

loggers = ['root']
handlers = []
formatters = []


class ConfigObjToConfigParser(configobj.ConfigObj):
    """Output a ``ConfigParser`` format from a ``ConfigObj`` input"""

    def _get_triple_quote(self, value):
        """Return the format string to format a multi-lines value

        In:
          - ``value`` -- the multi-lines value to format

        Return:
          - format object
        """
        # ``self.__mod__()`` will be called by the ``%``  operation
        return self

    def __mod__(self, value):
        """Format a multi-line value

        In:
          - ``value`` -- the multi-lines value to format

        Return:
          - the formatted multi-lines value
        """
        return '\n '.join(s.strip() for s in value.split('\n'))


apps_log_conf = ConfigObjToConfigParser({
    'loggers': {'keys': ''},
    'handlers': {'keys': ''},
    'formatters': {'keys': ''},
    'logger_root': {'handlers': ''}
}, list_values=False, indent_type='')


def configure(log_conf, app_name=None):
    """Merge all the applications logging configurations

    In:
      - ``log_conf`` -- an application logging configuration
      - ``app_name`` -- the name of the application
    """
    app_default_conf = configobj.ConfigObj(indent_type='', interpolation=False)

    if app_name:
        # The default application configuration
        app_default = {
            'logger_app_' + app_name: {
                'qualname': 'nagare.application.' + app_name,
                'level': 'INFO',
                'handlers': 'app_' + app_name,
                'propagate': '1'
            },

            'handler_app_' + app_name: {
                'class': 'StreamHandler',
                'formatter': 'app_' + app_name,
                'args': '(sys.stderr,)'
            },

            'formatter_app_' + app_name: {
                'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
            }
        }

        # Start with the default configuration
        app_default_conf.merge(configobj.ConfigObj(app_default, indent_type='', interpolation=False))

        # Create a dedicated logger, handler and formatter for the application
        loggers.append('app_' + app_name)
        handlers.append('app_' + app_name)
        formatters.append('app_' + app_name)

        # Create the list of all the handlers of the dedicated logger
        handler = log_conf.get('logger', {}).pop('handlers', None)
        if handler:
            app_default_conf['logger_app_' + app_name]['handlers'] += (', ' + handler)

        logger_sections = [values for (section, values) in log_conf.items() if section.startswith('logger_')]

        for logger_section in logger_sections:
            if ('qualname' in logger_section) and logger_section['qualname'].startswith('.'):
                logger_section['qualname'] = ('nagare.application.' + app_name + logger_section['qualname']).rstrip('.')

        # Use the generic 'logger', 'handler' and 'formatter' sections to
        # configure the dedicated logger, handler and formatter
        for section in ('logger', 'handler', 'formatter'):
            log_conf[section + '_app_' + app_name] = log_conf.pop(section, {})

    # Merge the loggers list
    keys = log_conf.get('loggers', {}).get('keys')
    if keys:
        if isinstance(keys, basestring):
            keys = [keys]
        loggers.extend(keys)

    # Merge the handlers list
    keys = log_conf.get('handlers', {}).get('keys')
    if keys:
        if isinstance(keys, basestring):
            keys = [keys]
        handlers.extend(keys)

    # Merge the formatters list
    keys = log_conf.get('formatters', {}).get('keys')
    if keys:
        if isinstance(keys, basestring):
            keys = [keys]
        formatters.extend(keys)

    # Merge the application configuration to the default configuration
    log_conf = configobj.ConfigObj(log_conf, interpolation=False)
    app_default_conf.merge(log_conf)

    # Merge the resulting configuration to the global configuration
    apps_log_conf.merge(app_default_conf)


def activate():
    """Use the merging of all the logging configurations to configure the Python logging system
    """
    logging.addLevelName(10000, 'NONE')

    apps_log_conf['loggers']['keys'] = ','.join(set(loggers))
    apps_log_conf['handlers']['keys'] = ','.join(set(handlers))
    apps_log_conf['formatters']['keys'] = ','.join(set(formatters))

    log_conf = StringIO.StringIO('\n'.join(apps_log_conf.write()))
    logging.config.fileConfig(log_conf)



#!/usr/bin/env python

# Copyright 2006 James Tauber and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import compiler
from compiler import ast
import os
import copy

class Klass:

    klasses = {}

    def __init__(self, name):
        self.name = name
        self.klasses[name] = self
        self.functions = set()

    def set_base(self, base_name):
        self.base = self.klasses.get(base_name)

    def add_function(self, function_name):
        self.functions.add(function_name)


class TranslationError(Exception):
    def __init__(self, message, node):
        self.message = "line %s:\n%s\n%s" % (node.lineno, message, node)

    def __str__(self):
        return self.message


class Translator:

    def __init__(self, module_name, mod, output):

        if module_name:
            self.module_prefix = module_name + "_"
        else:
            self.module_prefix = ""
        self.imported_modules = set()
        self.top_level_functions = set()
        self.top_level_classes = set()
        self.top_level_vars = set()
        self.output = output
        self.imported_classes = {}
        self.method_imported_globals = set()

        if mod.doc:
            print >>self.output, mod.doc

        for child in mod.node:
            if isinstance(child, ast.Function):
                self.top_level_functions.add(child.name)
            elif isinstance(child, ast.Class):
                self.top_level_classes.add(child.name)

        for child in mod.node:
            if isinstance(child, ast.Function):
                self._function(child)
            elif isinstance(child, ast.Class):
                self._class(child)
            elif isinstance(child, ast.Import):
                self.imported_modules.add(child.names[0][0])
            elif isinstance(child, ast.From):
                self.imported_modules.add(child.modname)
                self._from(child)
            elif isinstance(child, ast.Discard):
                self._discard(child, None)
            elif isinstance(child, ast.Assign):
                self._assign(child, None, True)
            else:
                raise TranslationError("unsupported type (in __init__)", child)


    def _function(self, node):
        function_name = self.module_prefix + node.name
        function_args = "(" + ", ".join(node.argnames) + ")"
        print >>self.output, "function %s%s {" % (function_name, function_args)

        if node.doc:
            print >>self.output, node.doc

        else:
            for child in node.code:
                self._stmt(child, None)

        print >>self.output, "}"
        print >>self.output, "\n"


    def _return(self, node, current_klass):
        expr = self.expr(node.value, current_klass)
        if expr != "null":
            print >>self.output, "    return " + expr + ";"
        else:
            print >>self.output, "    return;"


    def _break(self, node, current_klass):
        print >>self.output, "    break;"


    def _continue(self, node, current_klass):
        print >>self.output, "    continue;"


    def _callfunc(self, v, current_klass):
        if isinstance(v.node, ast.Name):
            if v.node.name in self.top_level_functions:
                call_name = self.module_prefix + v.node.name
            elif v.node.name in self.top_level_classes:
                call_name = self.module_prefix + v.node.name
            elif self.imported_classes.has_key(v.node.name):
                call_name = self.imported_classes[v.node.name] + '_' + v.node.name
            elif v.node.name == "callable":
                call_name = "pyjslib_isFunction"
            elif v.node.name == "map":
                call_name = "pyjslib_map"
            elif v.node.name == "filter":
                call_name = "pyjslib_filter"
            elif v.node.name == "dir":
                call_name = "pyjslib_dir"
            elif v.node.name == "getattr":
                call_name = "pyjslib_getattr"
            elif v.node.name == "hasattr":
                call_name = "pyjslib_hasattr"
            elif v.node.name == "int":
                call_name = "pyjslib_int"
            elif v.node.name == "str":
                call_name = "pyjslib_str"
            elif v.node.name == "range":
                call_name = "pyjslib_range"
            elif v.node.name == "len":
                call_name = "pyjslib_len"
            else:
                call_name = v.node.name
            call_args = []
        elif isinstance(v.node, ast.Getattr):
            attr_name = v.node.attrname
            if isinstance(v.node.expr, ast.Name):
                call_name = self._name2(v.node.expr, current_klass, attr_name)
                call_args = []
            elif isinstance(v.node.expr, ast.Getattr):
                call_name = self._getattr2(v.node.expr, current_klass, attr_name)
                call_args = []
            elif isinstance(v.node.expr, ast.CallFunc):
                call_name = self._callfunc(v.node.expr, current_klass) + "." + v.node.attrname
                call_args = []
            elif isinstance(v.node.expr, ast.Subscript):
                call_name = self._subscript(v.node.expr, current_klass) + "." + v.node.attrname
                call_args = []
            else:
                raise TranslationError("unsupported type (in _callfunc)", v.node.expr)
        else:
            raise TranslationError("unsupported type (in _callfunc)", v.node)

        for ch4 in v.args:
            arg = self.expr(ch4, current_klass)
            call_args.append(arg)

        return call_name + "(" + ", ".join(call_args) + ")"


    def _getattr(self, v):
        attr_name = v.attrname
        if isinstance(v.expr, ast.Name):
            obj = v.expr.name
            if obj == "self":
                obj = "this"
            elif self.imported_classes.has_key(obj):
                obj = "__" + self.imported_classes[obj] + '_' + obj
            elif obj in self.imported_modules:
                return obj + "_" + attr_name
            elif obj[0] == obj[0].upper():
                obj = "__" + self.module_prefix + obj
            return obj + "." + attr_name
        elif isinstance(v.expr, ast.Getattr):
            return self._getattr(v.expr) + "." + attr_name
        elif isinstance(v.expr, ast.Subscript):
            return self._subscript(v.expr, self.module_prefix) + "." + attr_name
        elif isinstance(v.expr, ast.CallFunc):
            return self._callfunc(v.expr, self.module_prefix) + "." + attr_name
        else:
            raise TranslationError("unsupported type (in _getattr)", v.expr)


    def _name(self, v):
        if v.name == "True":
            return "true"
        elif v.name == "False":
            return "false"
        elif v.name == "None":
            return "null"
        elif v.name == "self":
            return "this"
        elif v.name in self.top_level_classes:
            return self.module_prefix + v.name # @@@ should this be prefixed __?
        elif v.name in self.method_imported_globals:
            return self.module_prefix + v.name
        else:
            return v.name


    def _name2(self, v, current_klass, attr_name):
        obj = v.name

        if obj in self.method_imported_globals:
            call_name = self.module_prefix + obj + "." + attr_name
        elif self.imported_classes.has_key(obj):
            attr_str = ""
            if attr_name != "__init__":
                attr_str = ".prototype." + attr_name
            call_name = "__" + self.imported_classes[obj] + '_' + obj + attr_str + ".call"
        elif obj in self.imported_modules:
            call_name = obj + "_" + attr_name
        elif obj[0] == obj[0].upper():
            if attr_name == "__init__":
                call_name = "__" + self.module_prefix + obj + ".call"
            else:
                call_name = "__" + self.module_prefix + obj + ".prototype." + attr_name + ".call"
        else:
            if obj == "self":
                obj = "this"
            call_name = obj + "." + attr_name

        return call_name


    def _getattr2(self, v, current_klass, attr_name):
        if isinstance(v.expr, ast.Name):
            obj = v.expr.name
            if obj in self.imported_modules:
                if attr_name == "__init__":
                    call_name = "__" + obj + "_" + v.attrname + ".call"
                else:
                    call_name = "__" + obj + "_" + v.attrname + "." + attr_name + ".call"
            else:
                if obj == "self":
                    obj = "this"
                call_name = obj + "." + v.attrname + "." + attr_name
        elif isinstance(v.expr, ast.Getattr):
            call_name = self._getattr2(v.expr, current_klass, v.attrname + "." + attr_name)
        else:
            raise TranslationError("unsupport type (in _getattr2)", v.expr)

        return call_name


    def _class(self, node):

        class_name = self.module_prefix + node.name
        current_klass = Klass(node.name)

        for child in node.code:
            if isinstance(child, ast.Function):
                current_klass.add_function(child.name)

        if len(node.bases) == 0:
            base_class = ""
        elif len(node.bases) == 1:
            if isinstance(node.bases[0], ast.Name):
                if self.imported_classes.has_key(node.bases[0].name):
                    base_class = self.imported_classes[node.bases[0].name] + '_' + node.bases[0].name
                else:
                    base_class = self.module_prefix + node.bases[0].name
            elif isinstance(node.bases[0], ast.Getattr):
                base_class = self._name(node.bases[0].expr) + "_" + node.bases[0].attrname
            else:
                raise TranslationError("unsupported type (in _class)", node.bases[0])

            print >>self.output, "pyjs_extend(__" + class_name + ", __" + base_class + ");"
            current_klass.set_base(base_class)
        else:
            raise TranslationError("more than one base (in _class)", node)

        print >>self.output, "__" + class_name + '.prototype.__class__ = "' + class_name + '";'

        if "__init__" not in current_klass.functions:
            print >>self.output, "function " + class_name + "() {"
            print >>self.output, "    return new __" + class_name + "();"
            print >>self.output, "}"
            print >>self.output, "function __" + class_name + "() {"
            # call superconstructor
            if base_class:
                print >>self.output, "    __" + base_class + ".call(this);"
            print >>self.output, "}"

        for child in node.code:
            if isinstance(child, ast.Pass):
                pass
            elif isinstance(child, ast.Function):
                self._method(child, current_klass, class_name)
            elif isinstance(child, ast.Assign):
                self.classattr(child, current_klass)
            else:
                raise TranslationError("unsupported type (in _class)", child)

    def classattr(self, node, current_klass):
        self._assign(node, current_klass, True)

    def _method(self, node, current_klass, class_name):
        # reset global var scope
        self.method_imported_globals = set()

        arg_names = node.argnames

        if len(arg_names) == 0:
            raise TranslationError("first arg not 'self' (in _method)", node)
        if arg_names[0] != "self":
            raise TranslationError("first arg not 'self' (in _method)", node)
        function_args = "(" + ", ".join(arg_names[1:]) + ")"

        if node.name == "__init__":
            print >>self.output, "function " + class_name + function_args + " {"
            print >>self.output, "    return new __" + class_name + function_args + ";"
            print >>self.output, "}"
            print >>self.output, "function __" + class_name + function_args + " {"
        else:
            print >>self.output, "__" + class_name + ".prototype." + node.name + " = function" + function_args + " {"

        # default arguments
        if len(node.defaults):
            default_pos = len(arg_names) - len(node.defaults)
            for default_node in node.defaults:
                if isinstance(default_node, ast.Const):
                    default_value = self._const(default_node)
                elif isinstance(default_node, ast.Name):
                    default_value = self._name(default_node)
                else:
                    raise TranslationError("unsupported type (in _method)", default_node)

                default_name = arg_names[default_pos]
                default_pos += 1
                print >>self.output, "    if (typeof %s == 'undefined') %s=%s;" % (default_name, default_name, default_value)

        if node.doc:
            print >>self.output, node.doc

        else:
            for child in node.code:
                self._stmt(child, current_klass)

        if node.name == "__init__":
            print >>self.output, "}"
        else:
            print >>self.output, "};"


    def _stmt(self, node, current_klass):
        if isinstance(node, ast.Return):
            self._return(node, current_klass)
        elif isinstance(node, ast.Break):
            self._break(node, current_klass)
        elif isinstance(node, ast.Continue):
            self._continue(node, current_klass)
        elif isinstance(node, ast.Assign):
            self._assign(node, current_klass)
        elif isinstance(node, ast.AugAssign):
            self._augassign(node, current_klass)
        elif isinstance(node, ast.Discard):
            self._discard(node, current_klass)
        elif isinstance(node, ast.If):
            self._if(node, current_klass)
        elif isinstance(node, ast.For):
            self._for(node, current_klass)
        elif isinstance(node, ast.While):
            self._while(node, current_klass)
        elif isinstance(node, ast.Subscript):
            self._subscript_stmt(node, current_klass)
        elif isinstance(node, ast.Global):
            self._global(node, current_klass)
        elif isinstance(node, ast.Pass):
            pass
        else:
            raise TranslationError("unsupported type (in _stmt)", node)


    def _augassign(self, node, current_klass):
        v = node.node
        if isinstance(v, ast.Getattr):
            lhs = self._getattr(v)
        else:
            lhs = self._name(node.node)
        op = node.op
        rhs = self.expr(node.expr, current_klass)
        print >>self.output, "    " + lhs + " " + op + " " + rhs + ";"


    def _assign(self, node, current_klass, top_level = False):
        if len(node.nodes) != 1:
            raise TranslationError("not single node (in _assign)", node)
        v = node.nodes[0]
        if isinstance(v, ast.AssAttr):
            attr_name = v.attrname
            if isinstance(v.expr, ast.Name):
                obj = v.expr.name

                if obj in self.top_level_classes:
                    lhs = "__" + self.module_prefix + obj + "." + attr_name
                    lhs += " = __" + self.module_prefix + obj + ".prototype." + attr_name
                elif self.imported_classes.has_key(obj):
                    lhs = "__" + self.imported_classes[obj] + '_' + obj + "." + attr_name
                    lhs += " = __" + self.imported_classes[obj] + '_' +  obj + ".prototype." + attr_name
                else:
                    if obj == "self":
                        obj = "this"
                    lhs = "    " + obj + "." + attr_name

            elif isinstance(v.expr, ast.Getattr):
                lhs = "    " + self._getattr(v)
            elif isinstance(v.expr, ast.Subscript):
                lhs = "    " + self._subscript(v.expr, current_klass) + "." + attr_name
            else:
                raise TranslationError("unsupported type (in _assign)", v.expr)
            if v.flags == "OP_ASSIGN":
                op = "="
            else:
                raise TranslationError("unsupported flag (in _assign)", v)

        elif isinstance(v, ast.AssName):
            if top_level:
                if current_klass:
                    lhs = "__" + self.module_prefix + current_klass.name + "." + v.name
                    lhs += " = __" + self.module_prefix + current_klass.name + ".prototype." + v.name
                else:
                    self.top_level_vars.add(v.name)
                    lhs = "var " + self.module_prefix + v.name
            else:
                if v.name in self.method_imported_globals:
                    lhs = "    " + self.module_prefix + v.name
                else:
                    lhs = "    var " + v.name
            if v.flags == "OP_ASSIGN":
                op = "="
            else:
                raise TranslationError("unsupported flag (in _assign)", v)
        elif isinstance(v, ast.Subscript):
            if v.flags == "OP_ASSIGN":
                obj = self.expr(v.expr, current_klass)
                if len(v.subs) != 1:
                    raise TranslationError("must have one sub (in _assign)", v)
                idx = self.expr(v.subs[0], current_klass)
                value = self.expr(node.expr, current_klass)
                print >>self.output, "    " + obj + ".__setitem__(" + idx + ", " + value + ");"
                return
            else:
                raise TranslationError("unsupported flag (in _assign)", v)
        else:
            raise TranslationError("unsupported type (in _assign)", v)


        rhs = self.expr(node.expr, current_klass)
        print >>self.output, lhs + " " + op + " " + rhs + ";"


    def _discard(self, node, current_klass):
        if isinstance(node.expr, ast.CallFunc):
            expr = self._callfunc(node.expr, current_klass)
            print >>self.output, "    " + expr + ";"
        elif isinstance(node.expr, ast.Const):
            print >>self.output, node.expr.value
        else:
            raise TranslationError("unsupported type (in _discard)", node.expr)


    def _if(self, node, current_klass):
        for i in range(len(node.tests)):
            test, consequence = node.tests[i]
            if i == 0:
                keyword = "if"
            else:
                keyword = "else if"

            self._if_test(keyword, test, consequence, current_klass)

        if node.else_:
            keyword = "else"
            test = None
            consequence = node.else_

            self._if_test(keyword, test, consequence, current_klass)


    def _if_test(self, keyword, test, consequence, current_klass):

        if test:
            expr = self.expr(test, current_klass)

            print >>self.output, "    " + keyword + " (" + expr + ") {"
        else:
            print >>self.output, "    " + keyword + " {"

        if isinstance(consequence, ast.Stmt):
            for child in consequence.nodes:
                self._stmt(child, current_klass)
        else:
            raise TranslationError("unsupported type (in _if_test)", consequence)

        print >>self.output, "    }"


    def _from(self, node):
        for name in node.names:
            self.imported_classes[name[0]] = node.modname


    def _compare(self, node, current_klass):
        lhs = self.expr(node.expr, current_klass)

        if len(node.ops) != 1:
            raise TranslationError("only one ops supported (in _compare)", node)

        op = node.ops[0][0]
        rhs_node = node.ops[0][1]
        rhs = self.expr(rhs_node, current_klass)

        if op == "in":
            return rhs + ".__contains__(" + lhs + ")"
        elif op == "not in":
            return "!" + rhs + ".__contains__(" + lhs + ")"

        return "(" + lhs + " " + op + " " + rhs + ")"


    def _not(self, node, current_klass):
        expr = self.expr(node.expr, current_klass)

        return "!(" + expr + ")"

    def _or(self, node, current_klass):
        expr = "("+") || (".join([self.expr(child, current_klass) for child in node.nodes])+")"
        return expr

    def _and(self, node, current_klass):
        expr = "("+") && (".join([self.expr(child, current_klass) for child in node.nodes])+")"
        return expr

    def _for(self, node, current_klass):
        assign_name = ""
        assign_tuple = ""

        # based on Bob Ippolito's Iteration in Javascript code
        if isinstance(node.assign, ast.AssName):
            assign_name = node.assign.name
            if node.assign.flags == "OP_ASSIGN":
                op = "="
        elif isinstance(node.assign, ast.AssTuple):
            op = "="
            i = 0
            for child in node.assign:
                child_name = child.name
                if assign_name == "":
                    assign_name = "temp_" + child_name

                assign_tuple += """
                var %(child_name)s %(op)s %(assign_name)s.__getitem__(%(i)i);
                """ % locals()
                i += 1
        else:
            raise TranslationError("unsupported type (in _for)", node.assign)

        if isinstance(node.list, ast.Name):
            list_expr = self._name(node.list)
        elif isinstance(node.list, ast.Getattr):
            list_expr = self._getattr(node.list)
        elif isinstance(node.list, ast.CallFunc):
            list_expr = self._callfunc(node.list, current_klass)
        else:
            raise TranslationError("unsupported type (in _for)", node.list)

        lhs = "var " + assign_name
        iterator_name = "__" + assign_name

        print >>self.output, """
        var %(iterator_name)s = %(list_expr)s.__iter__();
        try {
            while (true) {
                %(lhs)s %(op)s %(iterator_name)s.next();
                %(assign_tuple)s
        """ % locals()
        for node in node.body.nodes:
            self._stmt(node, current_klass)
        print >>self.output, """
            }
        } catch (e) {
            if (e != StopIteration) {
                throw e;
            }
        }
        """ % locals()


    def _while(self, node, current_klass):
        test = self.expr(node.test, current_klass)
        print >>self.output, "    while (" + test + ") {"
        if isinstance(node.body, ast.Stmt):
            for child in node.body.nodes:
                self._stmt(child, current_klass)
        else:
            raise TranslationError("unsupported type (in _while)", node.body)
        print >>self.output, "    }"


    def _const(self, node):
        if isinstance(node.value, int):
            return str(node.value)
        elif isinstance(node.value, float):
            return str(node.value)
        elif isinstance(node.value, str):
            return '"' + node.value + '"'
        elif node.value is None:
            return "null"
        else:
            raise TranslationError("unsupported type (in _const)", node)

    def _unarysub(self, node, current_klass):
        return "-" + self.expr(node.expr, current_klass)

    def _add(self, node, current_klass):
        return self.expr(node.left, current_klass) + " + " + self.expr(node.right, current_klass)

    def _sub(self, node, current_klass):
        return self.expr(node.left, current_klass) + " - " + self.expr(node.right, current_klass)

    def _div(self, node, current_klass):
        return self.expr(node.left, current_klass) + " / " + self.expr(node.right, current_klass)

    def _mul(self, node, current_klass):
        return self.expr(node.left, current_klass) + " * " + self.expr(node.right, current_klass)

    def _mod(self, node, current_klass):
        return self.expr(node.left, current_klass) + " % " + self.expr(node.right, current_klass)

    def _invert(self, node, current_klass):
        return "("+"~" + self.expr(node.expr, current_klass)+")"

    def _bitand(self, node, current_klass):
        return "("+" & ".join([self.expr(child, current_klass) for child in node.nodes])+")"

    def _bitor(self, node, current_klass):
        return "("+" | ".join([self.expr(child, current_klass) for child in node.nodes])+")"

    def _subscript(self, node, current_klass):
        if node.flags == "OP_APPLY":
            if len(node.subs) == 1:
                return self.expr(node.expr, current_klass) + ".__getitem__(" + self.expr(node.subs[0], current_klass) + ")"
            else:
                raise TranslationError("must have one sub (in _subscript)", node)
        else:
            raise TranslationError("unsupported flag (in _subscript)", node)

    def _subscript_stmt(self, node, current_klass):
        if node.flags == "OP_DELETE":
            print >>self.output, "    " + self.expr(node.expr, current_klass) + ".__delitem__(" + self.expr(node.subs[0], current_klass) + ");"
        else:
            raise TranslationError("unsupported flag (in _subscript)", node)

    def _list(self, node, current_klass):
        return "new pyjslib_List([" + ", ".join([self.expr(x, current_klass) for x in node.nodes]) + "])"

    def _dict(self, node, current_klass):
        items = []
        for x in node.items:
            key = self.expr(x[0], current_klass)
            value = self.expr(x[1], current_klass)
            items.append("[" + key + ", " + value + "]")
        return "new pyjslib_Dict([" + ", ".join(items) + "])"

    def _tuple(self, node, current_klass):
        return "new pyjslib_Tuple([" + ", ".join([self.expr(x, current_klass) for x in node.nodes]) + "])"

    def _slice(self, node, current_klass):
        if node.flags == "OP_APPLY":
            lower = "null"
            upper = "null"
            if node.lower != None:
                lower = self.expr(node.lower, current_klass)
            if node.upper != None:
                upper = self.expr(node.upper, current_klass)
            return  "pyjslib_slice(" + self.expr(node.expr, current_klass) + ", " + lower + ", " + upper + ")"
        else:
            raise TranslationError("unsupported flag (in _slice)", node)

    def _global(self, node, current_klass):
        for name in node.names:
            self.method_imported_globals.add(name)

    def expr(self, node, current_klass):
        if isinstance(node, ast.Const):
            return self._const(node)
        # @@@ not sure if the parentheses should be here or in individual operator functions - JKT
        elif isinstance(node, ast.Mul):
            return " ( " + self._mul(node, current_klass) + " ) "
        elif isinstance(node, ast.Add):
            return " ( " + self._add(node, current_klass) + " ) "
        elif isinstance(node, ast.Sub):
            return " ( " + self._sub(node, current_klass) + " ) "
        elif isinstance(node, ast.Div):
            return " ( " + self._div(node, current_klass) + " ) "
        elif isinstance(node, ast.Mod):
            return self._mod(node, current_klass)
        elif isinstance(node, ast.UnarySub):
            return self._unarysub(node, current_klass)
        elif isinstance(node, ast.Not):
            return self._not(node, current_klass)
        elif isinstance(node, ast.Or):
            return self._or(node, current_klass)
        elif isinstance(node, ast.And):
            return self._and(node, current_klass)
        elif isinstance(node, ast.Invert):
            return self._invert(node, current_klass)
        elif isinstance(node, ast.Bitand):
            return self._bitand(node, current_klass)
        elif isinstance(node, ast.Bitor):
            return self._bitor(node, current_klass)
        elif isinstance(node, ast.Compare):
            return self._compare(node, current_klass)
        elif isinstance(node, ast.CallFunc):
            return self._callfunc(node, current_klass)
        elif isinstance(node, ast.Name):
            return self._name(node)
        elif isinstance(node, ast.Subscript):
            return self._subscript(node, current_klass)
        elif isinstance(node, ast.Getattr):
            return self._getattr(node)
        elif isinstance(node, ast.List):
            return self._list(node, current_klass)
        elif isinstance(node, ast.Dict):
            return self._dict(node, current_klass)
        elif isinstance(node, ast.Tuple):
            return self._tuple(node, current_klass)
        elif isinstance(node, ast.Slice):
            return self._slice(node, current_klass)
        else:
            raise TranslationError("unsupported type (in expr)", node)



import cStringIO

def translate(file_name, module_name):
    output = cStringIO.StringIO()
    mod = compiler.parseFile(file_name)
    t = Translator(module_name, mod, output)
    return output.getvalue()


class PlatformParser:
    def __init__(self, platform_dir = ""):
        self.platform_dir = platform_dir
        self.parse_cache = {}
        self.platform = ""

    def setPlatform(self, platform):
        self.platform = platform

    def parseModule(self, module_name, file_name):
        if self.parse_cache.has_key(file_name):
            mod = self.parse_cache[file_name]
        else:
            print "Importing " + module_name
            mod = compiler.parseFile(file_name)
            self.parse_cache[file_name] = mod

        platform_file_name = self.generatePlatformFilename(file_name)
        if self.platform and os.path.isfile(platform_file_name):
            mod = copy.deepcopy(mod)
            mod_override = compiler.parseFile(platform_file_name)
            self.merge(mod, mod_override)

        return mod

    def generatePlatformFilename(self, file_name):
        (module_name, extension) = os.path.splitext(os.path.basename(file_name))
        platform_file_name = module_name + self.platform + extension

        return os.path.join(os.path.dirname(file_name), self.platform_dir, platform_file_name)

    def merge(self, tree1, tree2):
        for child in tree2.node:
            if isinstance(child, ast.Function):
                self.replaceFunction(tree1, child.name, child)
            elif isinstance(child, ast.Class):
                self.replaceClassMethods(tree1, child.name, child)

        return tree1

    def replaceFunction(self, tree, function_name, function_node):
        # find function to replace
        for child in tree.node:
            if isinstance(child, ast.Function) and child.name == function_name:
                self.copyFunction(child, function_node)
                return
        raise TranslationError("function not found: " + function_name, function_node)

    def replaceClassMethods(self, tree, class_name, class_node):
        # find class to replace
        old_class_node = None
        for child in tree.node:
            if isinstance(child, ast.Class) and child.name == class_name:
                old_class_node = child
                break

        if not old_class_node:
            raise TranslationError("class not found: " + class_name, class_node)

        # replace methods
        for function_node in class_node.code:
            if isinstance(function_node, ast.Function):
                found = False
                for child in old_class_node.code:
                    if isinstance(child, ast.Function) and child.name == function_node.name:
                        found = True
                        self.copyFunction(child, function_node)
                        break

                if not found:
                    raise TranslationError("class method not found: " + class_name + "." + function_node.name, function_node)

    def copyFunction(self, target, source):
        target.code = source.code
        target.argnames = source.argnames
        target.defaults = source.defaults
        target.doc = source.doc


class AppTranslator:

    def __init__(self, library_dirs=["../library"], parser=None):
        self.extension = ".py"

        self.library_modules = []
        self.library_dirs = library_dirs

        if not parser:
            self.parser = PlatformParser()
        else:
            self.parser = parser

    def findFile(self, file_name):
        if os.path.isfile(file_name):
            return file_name

        for library_dir in self.library_dirs:
            full_file_name = os.path.join(os.path.dirname(__file__), library_dir, file_name)
            if os.path.isfile(full_file_name):
                return full_file_name

        raise Exception("file not found: " + file_name)

    def translate(self, module_name, is_app=True):

        if module_name not in self.library_modules:
            self.library_modules.append(module_name)

        file_name = self.findFile(module_name + self.extension)

        if is_app:
            module_name_translated = ""
        else:
            module_name_translated = module_name

        output = cStringIO.StringIO()

        mod = self.parser.parseModule(module_name, file_name)
        t = Translator(module_name_translated, mod, output)
        module_str = output.getvalue()

        imported_modules_str = ""
        for module in t.imported_modules:
            if module not in self.library_modules:
                imported_modules_str += self.translate(module, False)

        return imported_modules_str + module_str

    def translateLibraries(self, library_modules=[]):
        self.library_modules = library_modules

        imported_modules_str = ""
        for library in self.library_modules:
            imported_modules_str += self.translate(library, False)

        return imported_modules_str


if __name__ == "__main__":
    import sys
    file_name = sys.argv[1]
    if len(sys.argv) > 2:
        module_name = sys.argv[2]
    else:
        module_name = None
    print translate(file_name, module_name),



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

import os
import csv
import operator

from elixir import setup_all, create_all, drop_all, session
from elixir import Entity, Field, Unicode, has_many, belongs_to
from sqlalchemy import MetaData

from nose import with_setup
from paste.fixture import TestApp

from nagare.namespaces import xml
from nagare import presentation
from nagare import component
from nagare import wsgi
from nagare import local
from nagare.sessions.memory_sessions import SessionsWithPickledStates

from exceptions import Exception


def create_FixtureApp(app):
    local.worker = local.Process()
    local.request = local.Process()

    app = wsgi.create_WSGIApp(app)
    app.set_sessions_manager(SessionsWithPickledStates())
    app.start()

    return TestApp(app)


__metadata__ = MetaData()
__metadata__.bind = "sqlite:///:memory:"
__metadata__.bind.echo = False


# Setup / Teardown functions
def setup():
    setup_all()


def setup_func():
    create_all()


def teardown_func():
    session.expunge_all()
    drop_all()


class Language(Entity):
    id = Field(Unicode(50), primary_key=True)
    label = Field(Unicode(50))


class App1:
    def __init__(self):
        self.languages = []

    def add_language(self, language):
        self.languages.append(language)


@presentation.render_for(App1)
def render_app1(self, h, *args):
    return h.helloWorlds([h.helloWorld(language.label, {'language': language.id}) for language in self.languages])


@with_setup(setup_func, teardown_func)
def test_1():
    """ database - simple test with sqlalchemy/elixir """
    Language(id=u"english", label=u"hello world")
    session.flush()

    language = Language.query.all()[0]
    assert language.id == u"english"
    assert language.label == u"hello world"

    Language(id=u"french", label=u"bonjour monde")
    session.flush()

    language = Language.query.all()[1]
    assert language.id == u"french"
    assert language.label == u"bonjour monde"


@with_setup(setup_func, teardown_func)
def test_2():
    """ database - simple test with sqlalchemy/elixir unicode test """
    filePath = os.path.join(os.path.dirname(__file__), 'helloworld.csv')
    reader = csv.reader(open(filePath, 'r'))

    res = {}
    for row in reader:
        res[row[0].decode('utf-8')] = row[1].decode('utf-8')
        Language(id=row[0].decode('utf-8'), label=row[1].decode('utf-8'))
    session.flush()

    for language in Language.query.all():
        assert language.label == res[language.id]


@with_setup(setup_func, teardown_func)
def test_3():
    """ database - test with elixir & framework render method """
    app = App1()

    filePath = os.path.join(os.path.dirname(__file__), 'helloworld.csv')
    reader = csv.reader(open(filePath, 'r'))

    res = {}
    for row in reader:
        res[row[0].decode('utf-8')] = row[1].decode('utf-8')
        Language(id=row[0].decode('utf-8'), label=row[1].decode('utf-8'))
    session.flush()

    for language in Language.query.all():
        app.add_language(language)

    h = xml.Renderer()
    xmlToTest = component.Component(app).render(h).write_xmlstring(xml_declaration=True, pretty_print=True).strip()

    f = open(os.path.join(os.path.dirname(__file__), 'test_xmlns_1.xml'))
    xmlToCompare = f.read()
    f.close()

    assert xmlToTest == xmlToCompare


class Father(Entity):
    name = Field(Unicode(100))
    has_many('children', of_kind='Child')


class Child(Entity):
    name = Field(Unicode(100))
    belongs_to('father', of_kind='Father')


@with_setup(setup_func, teardown_func)
def test4():
    """ database - test children relation with sqlalchemy/elixir """

    f = Father(name=u"Father")

    c1 = Child(name=u"Child1")
    c2 = Child(name=u"Child2")

    f.children.append(c1)
    f.children.append(c2)
    assert f is c1.father
    assert f is c2.father
    session.flush()

    assert reduce(operator.__and__, [(elt.father.id == f.id) for elt in Child.query.all()])


def setup_func_2():
    create_all()
    Father(name=u"Charles Ingalls")
    session.flush()


class My_database_app():

    def __init__(self):
        self.father = Father.get_by(name=u'Charles Ingalls')

    def add_child(self, name):
        child = Child(name=name)
        self.father.children.append(child)
        if "Ingalls" not in name:
            raise Exception


@presentation.render_for(My_database_app, model='name')
def render(self, h, *args):
    h << h.div(self.father.name)
    return h.root


@presentation.render_for(My_database_app)
def render_database_app(self, h, comp, *args):
    h << h.div(self.father.name, id="father")
    h << [h.div(child.name, id="child") for child in self.father.children]

    h << h.a('Add Mary', id='add_mary').action(self.add_child, u'Mary Ingalls')
    h << h.a('Add Nancy', id='add_nancy').action(self.add_child, u'Nancy Oleson')
    h << h.a('Add Carrie', id='add_carrie').action(self.add_child, u'Carrie Ingalls')
    h << h.a('changes name', id='get_name').action(comp.becomes, model='name')

    return h.root


@with_setup(setup_func_2, teardown_func)
def test5():
    """ database - Get element in database """
    myApp = My_database_app
    app = create_FixtureApp(myApp)
    res = app.get('/')
    assert u"Charles Ingalls" in res


@with_setup(setup_func_2, teardown_func)
def test6():
    """ database - Add element in database """
    myApp = My_database_app
    app = create_FixtureApp(myApp)
    res = app.get('/')

    print res

    res = res.click(linkid="add_mary")
    assert Child.get_by(name=u"Mary Ingalls") is not None
    assert u"Charles Ingalls" in res
    assert u"Mary Ingalls" in res
    res = res.click(linkid="get_name")
    assert u"Charles Ingalls" in res


@with_setup(setup_func_2, teardown_func)
def test7():
    """ database - Add 2 elements in database """
    myApp = My_database_app
    app = create_FixtureApp(myApp)
    res = app.get('/')
    res = res.click(linkid="add_mary")
    assert Child.get_by(name=u"Mary Ingalls") is not None
    assert u"Charles Ingalls" in res
    assert u"Mary Ingalls" in res
    res = res.click(linkid="add_carrie")
    assert Child.get_by(name=u"Carrie Ingalls") is not None
    assert u"Charles Ingalls" in res
    assert u"Carrie Ingalls" in res
    assert u"Mary Ingalls" in res
    res = res.click(linkid="get_name")
    assert u"Charles Ingalls" in res


@with_setup(setup_func_2, teardown_func)
def test8():
    """ database - Add 2 elements in database and test rollback on Exception """
    myApp = My_database_app
    app = create_FixtureApp(myApp)
    res = app.get('/')
    res = res.click(linkid="add_mary")
    assert Child.get_by(name=u"Mary Ingalls") is not None
    assert u"Charles Ingalls" in res
    assert u"Mary Ingalls" in res

    try:
        res = res.click(linkid="add_nancy")
    except:
        res = app.get('/')
    assert Child.get_by(name=u"Nancy Oleson") is None
    assert u"Charles Ingalls" in res
    assert u"Nancy Oleson" not in res

    res = res.click(linkid="add_carrie")
    assert Child.get_by(name=u"Carrie Ingalls") is not None
    assert u"Charles Ingalls" in res
    assert u"Carrie Ingalls" in res
    assert u"Mary Ingalls" in res
    res = res.click(linkid="get_name")
    assert u"Charles Ingalls" in res



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""The ``create-rules`` administrative command

Generate the rewrite rules for the `apache <http://httpd.apache.org/>`_,
`lighttpd <http://www.lighttpd.net/>`_ or `nginx <http://nginx.net/>`_ http
servers, to let them serve the static contents of the launched applications,
instead of the applications themselves.
"""

import os
import pkg_resources

from nagare.admin import util, command


def set_options(optparser):
    """Register the possible options

    They are:

      - ``-a`` or ``--apache`` to generate the apache rules (default)
      - ``-l`` or ``--lighttpd`` to generate the lighttpd rules
      - ``-n`` or ``--nginx`` to generate the nginx rules

    In:
      - ``optparser`` -- the options parser
    """
    optparser.usage += ' [application ...]'

    optparser.add_option('-a', '--apache', action='store_const', const=generate_apache_rules, dest='generate', help='generate apache rules', default=generate_apache_rules)
    optparser.add_option('-l', '--lighttpd', action='store_const', const=generate_lighttpd_rules, dest='generate', help='generate lighttpd rules')
    optparser.add_option('-n', '--nginx', action='store_const', const=generate_nginx_rules, dest='generate', help='generate nginx rules')


def create_rules(app_names, error):
    """Generate the rewrite rules for the given registered applications

    In:
      - ``app_names`` -- list of the applications, read from the command line
      - ``error`` -- the function to call in case of configuration errors

    Return:
      - list of tuples (application, directory of its static contents)
    """
    # If no applications is given on the command line, generate the rules
    # for all the registered applications
    if not app_names:
        app_names = [entry.name for entry in pkg_resources.iter_entry_points('nagare.applications')]

    package = pkg_resources.Requirement.parse('nagare')
    static = pkg_resources.resource_filename(package, 'static')

    apps = [('nagare', static)]  # Initialize the result tuple with the static contents of the framework

    for app_name in app_names:
        (cfgfile, app, dist, aconf) = util.read_application(app_name, error)

        static = aconf['application'].get('static', os.path.join(dist.location, 'static') if dist else None)

        if static and os.path.isdir(static):
            apps.append((aconf['application']['name'], static))

    return sorted(apps, key=lambda x: len(x[0]))


def generate_lighttpd_rules(app_names, error):
    """Generate the lighttpd rewrite rules for the given registered applications

    In:
      - ``app_names`` -- list of the applications, read from the command line
      - ``error`` -- the function to call in case of configuration errors

    Return:
      - yield the lighttpd configuration file fragment
    """
    yield 'url.rewrite = ('
    for rule in create_rules(app_names, error):
        yield '  "^/static/%s/(.*)" => "%s/$1",' % rule
    yield ')'


def generate_nginx_rules(app_names, error):
    """Generate the nginx rewrite rules for the given registered applications

    In:
      - ``app_names`` -- list of the applications, read from the command line
      - ``error`` -- the function to call in case of configuration errors

    Return:
      - yield the nginx configuration file fragment
    """
    for (app_name, static) in create_rules(app_names, error):
        yield 'location /static/%s/ {' % app_name
        yield '  alias %s/;' % static
        yield '}'
        yield ''


def generate_apache_rules(app_names, error):
    """Generate the apache rewrite rules for the given registered applications

    In:
      - ``app_names`` -- list of the applications, read from the command line
      - ``error`` -- the function to call in case of configuration errors

    Return:
      - yield the apache configuration file fragment
    """
    yield 'RewriteEngine On'
    for rule in create_rules(app_names, error):
        yield 'RewriteRule ^/static/%s/(.*)$ %s/$1 [L]' % rule


def run(parser, options, args):
    print '\n'.join(options.generate(args, parser.error))


# ---------------------------------------------------------------------------

class CreateRules(command.Command):
    desc = 'Create the rewrite rules'

    set_options = staticmethod(set_options)
    run = staticmethod(run)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Simple form based authentication manager

The id and password of the user are first searched into the parameters of
the request. So, first, set a form with the fields names ``__ac_name``
and ``__ac_password`` (the prefix ``__ac`` is configurable).

Then the user id and the password are automatically kept into a cookie,
sent back on each request by the browser.

.. warning::

   This simple authentication manager keeps the user id and password in
   clear into the cookie. So this authentication manager is as secure as
   the HTTP basic authentication.
"""
from base64 import b64encode, b64decode

import webob

from nagare import security
from nagare.security import basic_auth


class Authentication(basic_auth.Authentication):
    """Simple form based authentication"""

    def __init__(self, prefix='__ac', key=None, max_age=None,
                 path='/', domain=None, secure=None, httponly=False,
                 comment=None, expires=None, overwrite=False,
                 realm=None):
        """Initialization

        In:
          - ``prefix`` -- prefix of the names of the user id and password fields
            into the form
          - ``realm`` -- is the form based authentication completed by a
            basic HTTP authentication ?
          - all the other keyword parameters are passed to the ``set_cookie()``
            method of the ``WebOb`` response object
            (see http://pythonpaste.org/webob/reference.html#id5)
        """
        super(Authentication, self).__init__(realm)
        self.prefix = prefix

        self.key = key or prefix
        self.max_age = max_age
        self.path = path
        self.domain = domain
        self.secure = secure
        self.httponly = httponly
        self.comment = comment
        self.expires = expires
        self.overwrite = overwrite

    def get_ids_from_params(self, params):
        """Search the data associated with the connected user into the request
        parameter

        In:
          - ``params`` -- the request parameters

        Return:
          - A tuple with the id of the user and its password
        """
        return params.get(self.prefix + '_name'), params.get(self.prefix + '_password')

    def cookie_decode(self, cookie):
        """Decode the data of the user cookie

        In:
          - ``cookie`` -- the data of the user cookie

        Return:
          - A list with the id of the user and its password
        """
        try:
            return [b64decode(s).decode('utf-8') for s in cookie.split(':')]
        except (TypeError, UnicodeDecodeError):
            # In case there's a problem when decoding the authentication cookie, fall back to unauthenticated
            return None, None

    def get_ids_from_cookie(self, cookies):
        """Search the data associated with the connected user into the cookies

        In:
          - ``cookies`` -- cookies dictionary

        Return:
          - A list with the id of the user and its password
        """
        data = cookies.get(self.key)
        if not data:
            return None, None

        return self.cookie_decode(data)

    def cookie_encode(self, *ids):
        """Encode the data of the user cookie

        In:
          - ``ids`` -- a tuple of data to put into the cookie

        Return:
          - the data to put into the user cookie
        """
        return ':'.join(b64encode(value.encode('utf-8')) for value in ids)

    def _get_ids(self, request, response):
        """Return the data associated with the connected user

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object

        Return:
          - A list with the id of the user and its password
        """
        # First, search into the request parameters
        ids = self.get_ids_from_params(request.params)
        if not all(ids):
            # Second, search into the cookies
            ids = self.get_ids_from_cookie(request.cookies)
            if not all(ids) and self.realm:
                # Third, if a realm is set, look into the basic authentication header
                ids = super(Authentication, self)._get_ids(request, response)

        return ids

    def set_user_id(self, user, id, password):
        """Set the credentials of the user

        In:
          - ``user`` -- the user
          - ``id`` -- the user id
          - ``password`` -- the user password
        """
        user.set_id(id, password)

    def end_rendering(self, request, response, session):
        """End of the request processing

        In:
          - ``request`` -- the request object
          - ``response`` -- the response object
          - ``session`` -- the session
        """
        user = security._get_user()

        if user:
            if not user.expired:
                response.set_cookie(
                    self.key, self.cookie_encode(*user.get_id()),
                    max_age=self.max_age,
                    path=self.path,
                    domain=self.domain,
                    secure=self.secure,
                    httponly=self.httponly,
                    comment=self.comment,
                    expires=self.expires,
                    overwrite=self.overwrite
                )
            else:
                response.delete_cookie(self.key, self.path, self.domain)

                if user.delete_session:
                    session.delete()

                location = user.logout_location
                if location is not None:
                    if not location.startswith(('http', '/')):
                        location = request.application_url + '/' + location

                    response.status = 301
                    response.location = location
                    response.body = ''

    def denies(self, detail):
        """Method called when a permission is denied

        In:
          - ``details`` -- a ``security.common.denial`` object
        """
        if self.realm:
            super(Authentication, self).denies(detail)

        raise webob.exc.HTTPForbidden(str(detail))

    def logout(self, location='', delete_session=True):
        """Deconnection of the current user

        Mark the user object as expired

        In:
          - ``location`` -- location to redirect to
          - ``delete_session`` -- is the session expired too ?
        """
        user = security.get_user()

        user.logout_location = location
        user.delete_session = delete_session
        user.expired = True



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

import time

import memcache

from nagare import local
from nagare.sessions import ExpirationError, common
from nagare.sessions.serializer import Pickle

KEY_PREFIX = 'nagare_%d_'


class Lock(object):
    def __init__(self, connection, lock_id, ttl, poll_time, max_wait_time):
        """Distributed lock in memcache

        In:
          - ``connection`` -- connection object to the memcache server
          - ``lock_id`` -- unique lock identifier
          - ``ttl`` -- session locks timeout, in seconds (0 = no timeout)
          - ``poll_time`` -- wait time between two lock acquisition tries, in seconds
          - ``max_wait_time`` -- maximum time to wait to acquire the lock, in seconds
        """
        self.connection = connection
        self.lock = (KEY_PREFIX + 'lock') % lock_id
        self.ttl = ttl
        self.poll_time = poll_time
        self.max_wait_time = max_wait_time

    def acquire(self):
        """Acquire the lock
        """
        t0 = time.time()
        while not self.connection.add(self.lock, 1, self.ttl) and (time.time() < (t0 + self.max_wait_time)):
            time.sleep(self.poll_time)

    def release(self):
        """Release the lock
        """
        self.connection.delete(self.lock)


class Sessions(common.Sessions):
    """Sessions manager for sessions kept in an external memcached server
    """
    spec = common.Sessions.spec.copy()
    spec.update(dict(
        host='string(default="127.0.0.1")',
        port='integer(default=11211)',
        ttl='integer(default=0)',
        lock_ttl='float(default=0.)',
        lock_poll_time='float(default=0.1)',
        lock_max_wait_time='float(default=5.)',
        min_compress_len='integer(default=0)',
        reset='boolean(default=True)',
        debug='boolean(default=False)',
        serializer='string(default="nagare.sessions.serializer:Pickle")'
    ))

    def __init__(
        self,
        host='127.0.0.1', port=11211,
        ttl=0,
        lock_ttl=0, lock_poll_time=0.1, lock_max_wait_time=5,
        min_compress_len=0,
        reset=False,
        debug=True,
        serializer=None,
        **kw
    ):
        """Initialization

        In:
          - ``host`` -- address of the memcache server
          - ``port`` -- port of the memcache server
          - ``ttl`` -- sessions and continuations timeout, in seconds (0 = no timeout)
          - ``lock_ttl`` -- session locks timeout, in seconds (0 = no timeout)
          - ``lock_poll_time`` -- wait time between two lock acquisition tries, in seconds
          - ``lock_max_wait_time`` -- maximum time to wait to acquire the lock, in seconds
          - ``min_compress_len`` -- data longer than this value are sent compressed
          - ``reset`` -- do a reset of all the sessions on startup ?
          - ``debug`` -- display the memcache requests / responses
          - ``serializer`` -- serializer / deserializer of the states
        """
        super(Sessions, self).__init__(serializer or Pickle, **kw)

        self.host = ['%s:%d' % (host, port)]
        self.ttl = ttl
        self.lock_ttl = lock_ttl
        self.lock_poll_time = lock_poll_time
        self.lock_max_wait_time = lock_max_wait_time
        self.min_compress_len = min_compress_len
        self.debug = debug

        if reset:
            self.flush_all()

    def set_config(self, filename, conf, error):
        """Read the configuration parameters

        In:
          - ``filename`` -- the path to the configuration file
          - ``conf`` -- the ``ConfigObj`` object, created from the configuration file
          - ``error`` -- the function to call in case of configuration errors
        """
        # Let's the super class validate the configuration file
        conf = super(Sessions, self).set_config(filename, conf, error)

        self.host = ['%s:%d' % (conf['host'], conf['port'])]

        for arg_name in (
            'ttl', 'lock_ttl', 'lock_poll_time', 'lock_max_wait_time',
            'min_compress_len', 'debug'
        ):
            setattr(self, arg_name, conf[arg_name])

        if conf['reset']:
            self.flush_all()

        return conf

    def _get_connection(self):
        """Get the connection to the memcache server

        Return:
          - the connection
        """
        # The connection objects are local to the workers
        connection = getattr(local.worker, 'memcached_connection', None)

        if connection is None:
            connection = memcache.Client(self.host, debug=self.debug)
            local.worker.memcached_connection = connection

        return connection

    def flush_all(self):
        """Delete all the contents in the memcached server
        """
        memcached = memcache.Client(self.host, debug=self.debug)
        memcached.flush_all()

    def get_lock(self, session_id):
        """Retrieve the lock of a session

        In:
          - ``session_id`` -- session id

        Return:
          - the lock
        """
        connection = self._get_connection()
        return Lock(connection, session_id, self.lock_ttl, self.lock_poll_time, self.lock_max_wait_time)

    def create(self, session_id, secure_id, lock):
        """Create a new session

        In:
          - ``session_id`` -- id of the session
          - ``secure_id`` -- the secure number associated to the session
          - ``lock`` -- the lock of the session
        """
        self._get_connection().set_multi({
            'state': 0,
            'sess': (secure_id, None),
            '00000': {}
        }, self.ttl, KEY_PREFIX % session_id, self.min_compress_len)

    def delete(self, session_id):
        """Delete a session

        In:
          - ``session_id`` -- id of the session to delete
        """
        self._get_connection().delete((KEY_PREFIX + 'sess') % session_id)

    def fetch_state(self, session_id, state_id):
        """Retrieve a state with its associated objects graph

        In:
          - ``session_id`` -- session id of this state
          - ``state_id`` -- id of this state

        Return:
          - id of the latest state
          - secure number associated to the session
          - data kept into the session
          - data kept into the state
        """
        state_id = '%05d' % state_id
        session = self._get_connection().get_multi(('state', 'sess', state_id), KEY_PREFIX % session_id)

        if len(session) != 3:
            raise ExpirationError()

        last_state_id = session['state']
        secure_id, session_data = session['sess']
        state_data = session[state_id]

        return last_state_id, secure_id, session_data, state_data

    def store_state(self, session_id, state_id, secure_id, use_same_state, session_data, state_data):
        """Store a state and its associated objects graph

        In:
          - ``session_id`` -- session id of this state
          - ``state_id`` -- id of this state
          - ``secure_id`` -- the secure number associated to the session
          - ``use_same_state`` -- is this state to be stored in the previous snapshot?
          - ``session_data`` -- data to keep into the session
          - ``state_data`` -- data to keep into the state
        """
        if not use_same_state:
            self._get_connection().incr((KEY_PREFIX + 'state') % session_id)

        self._get_connection().set_multi({
            'sess': (secure_id, session_data),
            '%05d' % state_id: state_data
        }, self.ttl, KEY_PREFIX % session_id, self.min_compress_len)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""The ``serve`` administrative command

Launch one or more applications.
"""

import os

import pkg_resources
import configobj

from nagare import config, log
from nagare.admin import reloader, util, reference, command

# ---------------------------------------------------------------------------

# The default publisher configuration
# -----------------------------------

publisher_options_spec = {
    'publisher': dict(
        host='string(default="127.0.0.1")',  # Listen only on the loopback interface
        port='integer(default=-1)'  # The default port depends of the publisher used
    ),

    'reloader': dict(
        activated='boolean(default=False)',  # No automatic reload
        interval='integer(default=1)',
    ),
}


# ---------------------------------------------------------------------------

def set_options(optparser):
    optparser.usage += ' [application ...]'

    optparser.add_option('-c', '--conf', dest='conf', help='configuration file')
    optparser.add_option('--host', action='store', type='string', help='Name of the interface to listen on ("0.0.0.0" to listen on all the interfaces)')
    optparser.add_option('-p', '--port', action='store', type='int', dest='port', help='Port to listen on')
    optparser.add_option('-d', '--debug', action='store_const', const='on', dest='debug', help='Activation of the web error page')
    optparser.add_option('--reload', action='store_const', const=True, dest='reload', help='Restart the application when a source file is changed')


def read_publisher_options(parser, options):
    """Read the configuration file for the publisher

    This configuration file is given with the ``-c``or ``--config`` option

    In:
      - ``parser`` -- the ``optparse.OptParser`` object used to parse the configuration file
      - ``options`` -- options in the command line

    Return:
      - a ``ConfigObj`` with the publisher parameters
    """
    if options.conf and not os.path.isfile(options.conf):
        parser.error('Configuration file "%s" doesn\'t exist' % options.conf)

    configspec = configobj.ConfigObj(publisher_options_spec)

    if options.conf:
        configspec.merge({'here': 'string(default="%s")' % os.path.abspath(os.path.dirname(options.conf))})

    choices = ', '. join(['"%s"' % entry.name for entry in pkg_resources.iter_entry_points('nagare.publishers')])
    configspec.merge({'publisher': {'type': 'option(%s, default="standalone")' % choices}})

    choices = ', '. join(['"%s"' % entry.name for entry in pkg_resources.iter_entry_points('nagare.sessions')])
    configspec.merge({'sessions': {'type': 'option(%s, default="standalone")' % choices}})

    conf = configobj.ConfigObj(options.conf, configspec=configspec, interpolation='Template')
    config.validate(options.conf, conf, parser.error)

    # The options in the command line overwrite the parameters read into the configuration file
    for (name, section, key) in (
        ('host', 'publisher', 'host'),
        ('port', 'publisher', 'port'),
        ('debug', 'publisher', 'debug'),
        ('reload', 'reloader', 'activated')
    ):
        option = getattr(options, name)
        if option is not None:
            conf[section][key] = option

    return conf


# ---------------------------------------------------------------------------

try:
    from weberror.evalexception import EvalException

    def debugged_app(app):
        return EvalException(app, xmlhttp_key='_a')
except ImportError:
    def debugged_app(app):
        return app


def create_wsgi_pipe(app, options, config_filename, config, error):
    """Wrap the application into one or more WSGI "middlewares" to create a WSGI pipe

    In:
      - ``app`` -- the application
      - ``options`` -- options in the command line
      - ``config_filename`` -- the path to the configuration file
      - ``config`` -- the ``ConfigObj`` object, created from the configuration file
      - ``error`` -- the function to call in case of configuration errors

    Return:
      - the wsgi pipe
    """
    if options.debug or config['application']['debug']:
        app = debugged_app(app)

    wsgi_pipe = config['application']['wsgi_pipe']
    if not wsgi_pipe:
        return app

    return reference.load_object(wsgi_pipe)[0](app, options, config_filename, config, error)


# ---------------------------------------------------------------------------

def get_file_from_root(root, path):
    """
    Return the path of a static content, from a filesystem root

    In:
      - ``root`` -- the path of the root
      - ``path`` -- the url path of the wanted static content

    Return:
      - the path of the static content
    """
    filename = os.path.join(root, path[1:])

    if not os.path.exists(filename) or os.path.isdir(filename):
        return None

    return filename


def get_file_from_package(package, path):
    """
    Return the path of a static content, from a setuptools package

    In:
      - ``package`` -- the setuptools package of a registered application
      - ``path`` -- the url path of the wanted static content

    Return:
      - the path of the static content
    """
    path = os.path.join('static', path[1:])

    if not pkg_resources.resource_exists(package, path) or pkg_resources.resource_isdir(package, path):
        return None

    return pkg_resources.resource_filename(package, path)


# ---------------------------------------------------------------------------

def run(parser, options, args):
    """Launch one or more applications

    In:
      - ``parser`` -- the ``optparse.OptParser`` object used to parse the configuration file
      - ``options`` -- options in the command lines
      - ``args`` -- arguments in the command lines

    The arguments are a list of names of registered applications
    or paths to application configuration files.
    """
    # If no argument are given, display the list of the registered applications
    if len(args) == 0:
        print 'Available applications:'

        # The applications are registered under the ``nagare.applications`` entry point
        applications = pkg_resources.iter_entry_points('nagare.applications')
        for name in sorted([application.name for application in applications]):
            print ' -', name
        return

    # Read the (optional) publisher configuration file
    pconf = read_publisher_options(parser, options)
    if pconf['reloader']['activated']:
        if 'nagare_reloaded' not in os.environ:
            return reloader.restart_with_monitor()

        filenames = pconf['reloader'].get('files', [])
        if isinstance(filenames, basestring):
            filenames = [filenames]
        filenames = filter(os.path.isfile, filenames)
        if options.conf:
            filenames.append(options.conf)

        watcher = reloader.install(pconf['reloader']['interval'], filenames, (pkg_resources.get_default_cache(),))
    else:
        watcher = None

    # Load the publisher
    publishers = dict([(entry.name, entry) for entry in pkg_resources.iter_entry_points('nagare.publishers')])
    t = pconf['publisher']['type']
    publisher = publishers[t].load()()

    # If no port is given, set the port number according to the used publisher
    if pconf['publisher']['port'] == -1:
        pconf['publisher']['port'] = publisher.default_port

    configs = []

    # Merge all the ``[logging]`` section of all the applications
    for cfgfile in args:
        # Read the configuration file of the application
        (cfgfile, app, dist, aconf) = util.read_application(cfgfile, parser.error)
        configs.append((cfgfile, app, dist, aconf))

        log.configure(aconf['logging'].dict(), aconf['application']['name'])

    # Configure the logging service
    log.activate()

    # Configure each application and register it to the publisher
    for (cfgfile, app, dist, aconf) in configs:
        # log.set_logger('nagare.application.'+aconf['application']['name'])

        if watcher:
            watcher.watch_file(aconf.filename)

        requirement = None if not dist else pkg_resources.Requirement.parse(dist.project_name)
        data_path = None if not requirement else pkg_resources.resource_filename(requirement, '/data')

        # Create the function to get the static contents of the application
        get_file = None
        static_path = aconf['application']['static']
        if static_path is not None and os.path.isdir(static_path):
            def get_file(path, static_path=static_path):
                return get_file_from_root(static_path, path)

        # Register the function to serve the static contents of the application
        static_url = publisher.register_static(aconf['application']['name'], get_file)

        # Load the sessions manager factory
        sessions_managers = dict([(entry.name, entry) for entry in pkg_resources.iter_entry_points('nagare.sessions')])
        conf = pconf['sessions'].dict()
        conf.update(aconf['sessions'].dict())
        t = conf.pop('type')

        sessions_manager = sessions_managers[t].load()()
        sessions_manager.set_config(options.conf, conf, parser.error)

        (app, metadatas) = util.activate_WSGIApp(
            app,
            cfgfile, aconf, parser.error,
            '' if not dist else dist.project_name,
            static_path, static_url,
            data_path,
            publisher,
            sessions_manager
        )

        # Register the application to the publisher
        publisher.register_application(
            aconf['application']['path'],
            aconf['application']['name'],
            app,
            create_wsgi_pipe(app, options, cfgfile, aconf, parser.error)
        )

    # Register the function to serve the static contents of the framework
    publisher.register_static(
        'nagare',
        lambda path, r=pkg_resources.Requirement.parse('nagare'): get_file_from_package(r, path)
    )

    # Launch all the applications
    publisher.serve(options.conf, pconf['publisher'], parser.error)


# ---------------------------------------------------------------------------

class Serve(command.Command):
    desc = 'Launch applications'

    set_options = staticmethod(set_options)
    run = staticmethod(run)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

import sys
import mimetypes

import pkg_resources


if sys.version_info < (2, 5, 2):
    print 'The version of Python must be 2.5.2 or more'
    sys.exit(-2)

# -----------------------------------------------------------------------------

pkg_resources.declare_namespace(__name__)

# -----------------------------------------------------------------------------

# Fix issue 5868 in Python 2.6.2 (http://bugs.python.org/issue5868)
mimetypes.init()



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""The HTTP multi-threaded publisher"""

import time

from paste import httpserver

from nagare import local
from nagare.publishers import common


class Publisher(common.Publisher):
    """The HTTP publisher"""

    # Possible configuration options
    # ------------------------------

    server_spec = dict(
        host='string(default=None)', port='integer(default=None)',
        server_version='string(default=None)', protocol_version='string(default=None)',
        daemon_threads='boolean(default=None)',
        socket_timeout='integer(default=None)',
        use_threadpool='boolean(default=None)', threadpool_workers='integer(default=None)'
    )

    threadpool_spec = dict(
        max_requests='integer(default=None)',  # threads are killed after this many requests
        hung_thread_limit='integer(default=None)',  # when a thread is marked "hung"
        kill_thread_limit='integer(default=None)',  # when you kill that hung thread
        dying_limit='integer(default=None)',  # seconds that a kill should take to go into effect (longer than this and the thread is a "zombie")
        spawn_if_under='integer(default=None)',  # spawn if there's too many hung threads
        max_zombie_threads_before_die='integer(default=None)',  # when to give up on the process
        hung_check_period='integer(default=None)'  # every 100 requests check for hung workers
    )

    spec = server_spec.copy()
    spec.update(threadpool_spec)

    def __init__(self):
        """Initialization
        """
        super(Publisher, self).__init__()

        local.worker = local.Thread()
        local.request = local.Thread()

    def serve(self, filename, conf, error):
        """Run the publisher

        In:
          - ``filename`` -- the path to the configuration file
          - ``conf`` -- the ``ConfigObj`` object, created from the configuration file
          - ``error`` -- the function to call in case of configuration errors
        """
        (host, port, conf) = self._validate_conf(filename, conf, error)

        # The publisher is a threaded server so call only once the ``on_new_process()`` method
        self.on_new_process()
        print time.strftime('%x %X -', time.localtime()),

        server_options = dict([(k, v) for (k, v) in conf.items() if k in self.server_spec])
        threadpool_options = dict([(k, v) for (k, v) in conf.items() if k in self.threadpool_spec])

        httpserver.serve(self.urls, host, port, threadpool_options=threadpool_options, **server_options)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""The ``serve-module`` administrative command

Launch a standalone (without any database connection) application, in debug mode.

No configuration file in read. The only possible parameters are given on the
command line and are (run ``nagare-admin serve-module`` to see them):

  - the host
  - the port
  - the web debug mode
  - the reload mode

Also no static contents can be served.

.. warning::
    Use this command in development mode or to quickly test a piece code.
    Don't run a production application with this command
"""

import os

import pkg_resources

from nagare import wsgi, log
from nagare.admin import reference, command, reloader

try:
    from weberror.evalexception import EvalException

    def debugged_app(app):
        return EvalException(app, xmlhttp_key='_a')
except ImportError:
    def debugged_app(app):
        return app


def get_file_from_package(package, path):
    """
    Return the path of a static content, located into a setuptools package

    In:
      - ``package`` -- the setuptools package of a registered application
      - ``path`` -- the url path of the wanted static content

    Return:
      - the path of the static content
    """
    path = os.path.join('static', path[1:])

    if not pkg_resources.resource_exists(package, path) or pkg_resources.resource_isdir(package, path):
        return None

    return pkg_resources.resource_filename(package, path)


def set_options(optparser):
    optparser.usage += ' <module_or_file> <name>'

    optparser.add_option('--host', action='store', type='string', default='127.0.0.1', help='Name of the interface to listen on ("0.0.0.0" to listen on all the interfaces)')
    optparser.add_option('-p', '--port', action='store', type='int', dest='port', default=8080, help='Port to listen on')
    optparser.add_option('--no-debug', action='store_const', const=False, default=True, dest='debug', help='Desactivation of the web error page')


def run(parser, options, args):
    """launch an object

    In:
      - ``parser`` -- the ``optparse.OptParser`` object used to parse the configuration file
      - ``options`` -- options in the command lines
      - ``args`` -- arguments in the command lines

    The unique argument is the path of the object to launch. The path syntax is described
    into the module ``nagare.admin.reference``. For example, ``/tmp/counter.py:Counter``
    is the path to the class ``Counter`` of the module ``tmp.counter.py``

    """
    if len(args) != 2:
        parser.error('Bad number of arguments')

    if 'nagare_reloaded' not in os.environ:
        return reloader.restart_with_monitor()

    # With the ``serve-module`` command, the automatic reloader is always activated
    reloader.install(excluded_directories=(pkg_resources.get_default_cache(),))

    # Load the object
    if os.path.sep in args[0]:
        path = 'file ' + args[0]
    else:
        path = 'python ' + args[0]
    app = reference.load_object(path)[0]

    # Wrap it into a WSGIApp
    app = wsgi.create_WSGIApp(app)

    # Always use the standalone publisher (Python HTTP server)
    publishers = dict([(entry.name, entry) for entry in pkg_resources.iter_entry_points('nagare.publishers')])
    publisher = publishers['standalone'].load()()

    wsgi_pipe = debugged_app(app) if options.debug else app
    publisher.register_application(args[0], args[1], app, wsgi_pipe)
    app.set_config('', {'application': {'redirect_after_post': False, 'name': args[1], 'always_html': True}}, None)
    app.set_publisher(publisher)

    # Always use the standalone session manager (in memory sessions)
    sessions_managers = dict([(entry.name, entry) for entry in pkg_resources.iter_entry_points('nagare.sessions')])
    sessions_manager = sessions_managers['standalone'].load()()
    app.set_sessions_manager(sessions_manager)

    # Set the application logger level to DEBUG
    log.configure({'logger': {'level': 'DEBUG'}}, args[1])
    log.activate()
    log.set_logger('nagare.application.' + args[1])

    # The static contents of the framework are served by the standalone server
    publisher.register_static('nagare', lambda path, r=pkg_resources.Requirement.parse('nagare'): get_file_from_package(r, path))

    # Launch the object
    publisher.serve(None, dict(host=options.host, port=options.port), None)


# ---------------------------------------------------------------------------

class Serve(command.Command):
    desc = 'Launch a python module'

    set_options = staticmethod(set_options)
    run = staticmethod(run)



# (c) 2005 Ian Bicking and contributors; written for Paste (http://pythonpaste.org)
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php

"""Utiliy module that watches for files modification and exits the current
process when this event happens.

This module is derivated from Ian Bicking's one to only exit if there is no
importation errors into the modified file.
"""

import os
import sys
import time
import platform
import threading
import traceback
import subprocess


def _turn_sigterm_into_systemexit():
    """
    Attempts to turn a SIGTERM exception into a SystemExit exception.
    """
    try:
        import signal
    except ImportError:
        return

    def handle_term(signo, frame):
        raise SystemExit

    signal.signal(signal.SIGTERM, handle_term)


def _quote_first_command_arg(arg):
    """
    There's a bug in Windows when running an executable that's
    located inside a path with a space in it.  This method handles
    that case, or on non-Windows systems or an executable with no
    spaces, it just leaves well enough alone.
    """
    if (sys.platform != 'win32') or (' ' not in arg):
        # Problem does not apply:
        return arg

    try:
        import win32api
    except ImportError:
        raise ValueError(
            "The executable %r contains a space, and in order to "
            "handle this issue you must have the win32api module "
            "installed" % arg)

    return win32api.GetShortPathName(arg)


def restart_with_monitor():
    while True:
        args = [_quote_first_command_arg(sys.executable)] + sys.argv
        new_environ = os.environ.copy()
        new_environ['nagare_reloaded'] = 'True'

        proc = None
        try:
            try:
                _turn_sigterm_into_systemexit()
                proc = subprocess.Popen(args, env=new_environ)
                exit_code = proc.wait()
                proc = None
            except KeyboardInterrupt:
                print '^C caught in monitor process'
                # if self.verbose > 1:
                #     raise
                return 1
        finally:
            if (proc is not None) and hasattr(os, 'kill') and (platform.system() != 'Windows'):
                import signal
                try:
                    os.kill(proc.pid, signal.SIGTERM)
                except (OSError, IOError):
                    raise

        if exit_code != 3:
            return exit_code

        print '-' * 20, 'Restarting', '-' * 20


class Monitor(object):
    def __init__(self, poll_interval, included_files, excluded_directories):
        self.poll_interval = poll_interval
        self.included_files = list(included_files)
        self.excluded_directories = [d + ('' if d.endswith(os.sep) else os.sep) for d in excluded_directories]

        self.mtimes = {}

    def watch_file(self, filename):
        self.included_files.append(filename)

    def periodic_reload(self):
        while not self.check_modifications():
            time.sleep(self.poll_interval)
        os._exit(3)

    def check_modification(self, filename):
        try:
            mtime = os.stat(filename).st_mtime
        except (OSError, IOError):
            return False

        r = (self.mtimes.setdefault(filename, mtime) < mtime)
        self.mtimes[filename] = mtime

        return r

    def check_modifications(self):
        for filename in self.included_files:
            if self.check_modification(filename):
                print >> sys.stderr, 'File %s changed; reloading...' % filename
                return True

        for m in sys.modules.values():
            if not hasattr(m, '__file__'):
                continue

            filename = m.__file__

            for directory in self.excluded_directories:
                if filename.startswith(directory):
                    continue

            if filename.endswith(('.pyc', '.pyo')) and os.path.exists(filename[:-1]):
                filename = filename[:-1]

            if self.check_modification(filename):
                try:
                    reload(m)
                    print >> sys.stderr, 'Module %s changed; reloading...' % filename
                    return True
                except Exception:
                    print >> sys.stderr, 'WARNING, module %s NOT reloaded' % filename
                    print >> sys.stderr, traceback.format_exc()

        return False


def install(poll_interval=1, included_files=(), excluded_directories=()):
    mon = Monitor(poll_interval, included_files, excluded_directories)
    t = threading.Thread(target=mon.periodic_reload)
    t.setDaemon(True)
    t.start()
    return mon
